#    Copyright (C) 2017 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

import sys
import ast
import re
import numpy as N

from .. import qtall as qt

def _(text, disambiguation=None, context="Import_FITS_HDF5"):
    return qt.QCoreApplication.translate(context, text, disambiguation)

def filterAttrsByName(attrs, name):
    """For compound datasets, attributes can be given on a per-column basis.
    This filters the attributes by the column name."""

    name = name.strip()
    attrsout = {}
    for a in attrs:
        # attributes with _dsname suffixes are copied
        if a[:4] == "vsz_" and a[-len(name)-1:] == "_"+name:
            attrsout[a[:-len(name)-1]] = attrs[a]
    return attrsout

def convertTextToSlice(slicetxt, numdims):
    """Convert a value like 0:1:3,:,::-1 to a tuple slice
    ((0,1,3), (None, None, None), (None, None, -1))
    or reduce dimensions such as :,3 -> ((None,None,None),3)

    Also checks number of dimensions (including reduced) is numdims.

    Return -1 on error
    """

    if slicetxt.strip() == '':
        return None

    slicearray = slicetxt.split(',')
    if len(slicearray) != numdims:
        # slice needs same dimensions as data
        return -1

    allsliceout = []
    for sliceap_idx, sliceap in enumerate(slicearray):
        sliceparts = sliceap.strip().split(':')

        if len(sliceparts) == 1:
            # reduce dimensions with single index
            try:
                allsliceout.append(int(sliceparts[0]))
            except ValueError:
                # invalid index
                return -1
        elif len(sliceparts) not in (2, 3):
            return -1
        else:
            sliceout = []
            for p in sliceparts:
                p = p.strip()
                if not p:
                    sliceout.append(None)
                else:
                    try:
                        sliceout.append(int(p))
                    except ValueError:
                        return -1
            if len(sliceout) == 2:
                sliceout.append(None)
            allsliceout.append(tuple(sliceout))

    allempty = True
    for s in allsliceout:
        if s != (None, None, None):
            allempty = False
    if allempty:
        return None

    return tuple(allsliceout)

def convertSliceToText(slice):
    """Convert tuple slice into text."""
    if slice is None:
        return ''
    out = []
    for spart in slice:
        if isinstance(spart, int):
            # single index
            out.append(str(spart))
            continue

        sparttxt = []
        for p in spart:
            if p is not None:
                sparttxt.append(str(p))
            else:
                sparttxt.append('')
        if sparttxt[-1] == '':
            del sparttxt[-1]
        out.append(':'.join(sparttxt))
    return ', '.join(out)

def applySlices(data, slices):
    """Given hdf/numpy dataset, apply slicing tuple to it and return data."""
    slist = []
    for s in slices:
        if isinstance(s, int):
            slist.append(s)
        else:
            slist.append(slice(*s))
            if s[2] is not None and s[2] < 0:
                # negative slicing doesn't work in h5py, so we
                # make a copy
                data = N.array(data)
    try:
        data = data[tuple(slist)]
    except (ValueError, IndexError):
        data = N.array([], dtype=N.float64)
    return data

class ConvertError(RuntimeError):
    pass

def convertFromBytes(s):
    """h5py often returns bytes instead of unicode.
    This decodes if in bytes
    """
    if sys.version_info[0] == 3:
        if isinstance(s, bytes):
            return s.decode('utf-8')
    return s

def convertDatasetToObject(data, slices):
    """Convert numpy/hdf dataset to suitable data for veusz.
    Raise ConvertError if cannot."""

    # lazily-loaded h5py
    try:
        from h5py import check_dtype
    except ImportError:
        # fallback if no h5py, e.g. only installed fits
        def check_dtype(vlen=None):
            return False

    if slices:
        data = applySlices(data, slices)

    try:
        kind = data.dtype.kind
    except TypeError:
        raise ConvertError(_("Could not get data type of dataset"))

    if kind in ('b', 'i', 'u', 'f'):
        data = N.array(data, dtype=N.float64)
        if data.ndim == 0:
            raise ConvertError(_("Dataset has no dimensions"))
        return data

    elif kind in ('S', 'a', 'U') or (
        kind == 'O' and check_dtype(vlen=data.dtype) is str):
        if hasattr(data, 'ndim') and data.ndim != 1:
            raise ConvertError(_("Text datasets must have 1 dimension"))

        strcnv = list(data)
        return strcnv

    raise ConvertError(_("Dataset has an invalid type"))

def getFITSHduNames(fitsfile):
    """Return list of names to give HDUs given a FITS file."""

    nameset = set()
    names = []

    for i, hdu in enumerate(fitsfile):
        name = hdu.name

        if not name:
            name = 'hdu%i' % i
            # just in case people start naming HDUs hduX...
            while name in nameset:
                name += '~'
        else:
            name = name.lower()

            # EXTVER distinguishes identical names
            if 'EXTVER' in hdu.header:
                name = '%s%i' % (name, hdu.header['EXTVER'])

            # prevent duplicates
            if name in nameset:
                i = 2
                while name+str(i) in nameset:
                    i += 1
                name += str(i)

        nameset.add(name)
        names.append(name)

    return names

def convertFITSDataFormat(fmt):
    """Convert FITS TFORM codes into:

    (code, nlen)

    code is 'invalid', 'numeric' or 'text'
    nlen is number of entries for column
    """

    # match the fits format text code [r]F[w[.d]]
    m = re.match(r'^([0-9]*)([A-Za-z])([0-9]*)(\.[0-9]+)?$', fmt)
    if not m:
        return (None, 'invalid', ())
    grps = m.groups()

    # length of array
    try:
        nlen = int(grps[0])
    except ValueError:
        nlen = 1
    fcode = grps[1].upper()
    width = grps[2]

    if fcode == 'A' and not width:
        # note: we can't handle 2d text arrays, so we handle as invalid
        code = 'text'
        # even though strings are N characters, they are handled as singles
        nlen = 1

    elif fcode in 'LXBIJKED':
        code = 'numeric'

    else:
        code = 'invalid'

    return code, nlen

def hduVeuszAttrs(hdu):
    """Get veusz-specific attributes from a HDU header.

    These use the VEUSZ keyword and have the format of
    KEY=VALUE
     or
    COLUMN: KEY=VALUE

    Returns (attrs, colattrs)

    Where attrs is HDU-specific and colattrs is column specific (dict
    of dicts)

    """

    attrs = {}
    colattrs = {}

    for k, v in hdu.header.items():
        if k.lower() == 'veusz':

            # match syntax [OPTIONAL COLUMN:] KEY=VALUE
            match = re.match(
                r'^(?:([a-zA-Z0-9_]+)[ ]*:)?[ ]*([a-zA-Z0-9_]+)[ ]*=[ ]*(.*)$',
                v)
            if not match:
                continue

            col, key, tval = match.groups()
            key = key.lower()
            tval = tval.strip()

            # convert to python type if possible
            try:
                val = ast.literal_eval(tval)
            except Exception:
                val = tval

            if col:
                col = col.lower()

                # column-specific key value
                if col not in colattrs:
                    colattrs[col] = {key: val}
                else:
                    colattrs[col][key] = val
            else:
                # hdu-specific key value
                attrs[key] = val

    return attrs, colattrs



#    Copyright (C) 2010 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Polar plot widget."""

import numpy as N
import fractions
import math

from .nonorthgraph import NonOrthGraph
from .axisticks import AxisTicks
from . import axis

from .. import qtall as qt
from .. import document
from .. import setting
from .. import utils

def _(text, disambiguation=None, context='Polar'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

def radianToAlign(a):
    # fraction of circle
    f = a / (2*math.pi)

    # slow, but ok here as we don't expect values much outside range
    while f < 0:
        f += 1
    while f >= 1:
        f -= 1

    if f == 0:
        return (-1, 1)
    elif f < 0.25:
        return (-1, 1)
    elif f == 0.25:
        return (0, 1)
    elif f < 0.5:
        return (1, 1)
    elif f == 0.5:
        return (1, 0)
    elif f < 0.75:
        return (1, -1)
    elif f == 0.75:
        return (0, -1)
    else:
        return (-1, -1)

class OldLine(setting.Line):
    '''Polar tick settings.'''

    def __init__(self, name):
        setting.Line.__init__(self, name, setnsmode='hide')

        self.add( setting.SettingBackwardCompat(
            'number', '../RadiiLine/number', None))
        self.add( setting.SettingBackwardCompat(
            'hidespokes', '../SpokeLine/length', None))
        self.add( setting.SettingBackwardCompat(
            'hideannuli', '../SpokeAnnuli/length', None))

class SpokeLine(setting.Line):
    '''Spokes in polar plot.'''

    def __init__(self, name, **args):
        setting.Line.__init__(self, name, **args)
        self.add( setting.Int(
            'number', 12,
            minval=1,
            descr=_('Number of spokes to use'),
            usertext=_('Number')) )
        self.get('color').newDefault('grey')

class RadiiLine(setting.Line):
    '''Radii in polar plot.'''

    def __init__(self, name, **args):
        setting.Line.__init__(self, name, **args)
        self.add( setting.Int(
            'number', 6,
            minval=1,
            descr=_('Number of radial ticks to aim for'),
            usertext=_('Number')) )
        self.get('color').newDefault('grey')

class TickLabel(axis.TickLabel):
    """For tick label."""
    def __init__(self, *args, **argsv):
        axis.TickLabel.__init__(self, *args, **argsv)
        self.remove('offset')
        self.remove('rotate')
        self.remove('hide')
        self.add( setting.Bool(
            'hideradial', False,
            descr=_('Hide radial labels'),
            usertext=_('Hide radial') ) )
        self.add( setting.Bool(
            'hidetangential', False,
            descr=_('Hide spoke labels'),
            usertext=_('Hide spokes') ) )

class Polar(NonOrthGraph):
    '''Polar plotter.'''

    typename='polar'
    allowusercreation = True
    description = _('Polar graph')

    @classmethod
    def addSettings(klass, s):
        '''Construct list of settings.'''
        NonOrthGraph.addSettings(s)

        s.add( setting.FloatOrAuto(
            'minradius', 'Auto',
            descr=_('Minimum value of radius'),
            usertext=_('Min radius')) )
        s.add( setting.FloatOrAuto(
            'maxradius', 'Auto',
            descr=_('Maximum value of radius'),
            usertext=_('Max radius')) )
        s.add( setting.Choice(
            'units',
            ('degrees', 'radians', 'fractions', 'percentages'),
            'degrees',
            descr=_('Angular units'),
            usertext=_('Units')) )
        s.add( setting.Choice(
            'direction',
            ('clockwise', 'anticlockwise'),
            'anticlockwise',
            descr=_('Angle direction'),
            usertext=_('Direction')) )
        s.add( setting.Choice(
            'position0',
            ('right', 'top', 'left', 'bottom'),
            'right',
            descr=_('Direction of 0 angle'),
            usertext=_('Position of 0°')) )
        s.add( setting.Bool(
            'log', False,
            descr=_('Logarithmic radial axis'),
            usertext=_('Log')) )

        s.add( TickLabel(
            'TickLabels',
            descr=_('Radial tick labels'),
            usertext=_('Radial tick labels')),
            pixmap='settings_axisticklabels' )

        s.add( OldLine('Tick') )

        s.add( SpokeLine(
            'SpokeLine',
            descr=_('Spoke line'),
            usertext=_('Spoke line')),
            pixmap='settings_axismajorticks' )

        s.add( RadiiLine(
            'RadiiLine',
            descr=_('Radii line'),
            usertext=_('Radii line')),
            pixmap='settings_contourline' )

        s.get('leftMargin').newDefault('1cm')
        s.get('rightMargin').newDefault('1cm')
        s.get('topMargin').newDefault('1cm')
        s.get('bottomMargin').newDefault('1cm')

    @property
    def userdescription(self):
        s = self.settings
        return _("'units=%s, direction=%s, log=%s") % (
            s.units, s.direction, str(s.log))

    def coordRanges(self):
        '''Get ranges of coordinates.'''

        angularrange = {
            'degrees': [0., 360.],
            'radians': [0., 2*math.pi],
            'fractions': [0., 1.],
            'percentages': [0., 100.],
        }[self.settings.units]

        return [
            [self._minradius, self._maxradius],
            angularrange
        ]

    def toPlotAngle(self, angles):
        """Convert one or more angles to angle on plot."""
        s = self.settings

        # unit conversions
        mult = {
            'degrees': math.pi/180,
            'radians': 1,
            'fractions': 2*math.pi,
            'percentages': 0.02*math.pi,
        }
        angles = angles * mult[s.units]

        # change direction
        if self.settings.direction == 'anticlockwise':
            angles = -angles
        # add offset
        angles -= {
            'right': 0, 'top': 0.5*math.pi, 'left': math.pi,
            'bottom': 1.5*math.pi}[self.settings.position0]
        return angles

    def toPlotRadius(self, radii):
        """Convert radii to a plot radii."""
        if self.settings.log:
            logmin = N.log(self._minradius)
            logmax = N.log(self._maxradius)
            r = (N.log(N.clip(radii, 1e-99, 1e99)) - logmin) / (
                logmax - logmin)
        else:
            r = (radii - self._minradius) / (
                self._maxradius - self._minradius)
        return N.where(r > 0., r, 0.)

    def graphToPlotCoords(self, coorda, coordb):
        '''Convert coordinates in r, theta to x, y.'''

        ca = self.toPlotRadius(coorda)
        cb = self.toPlotAngle(coordb)

        x = self._xc + ca * N.cos(cb) * self._xscale
        y = self._yc + ca * N.sin(cb) * self._yscale
        return x, y

    def drawFillPts(self, painter, extfill, cliprect,
                    ptsx, ptsy):
        '''Draw points for plotting a fill.'''
        pts = qt.QPolygonF()
        utils.addNumpyToPolygonF(pts, ptsx, ptsy)

        filltype = extfill.filltype
        if filltype == 'center':
            pts.append( qt.QPointF(self._xc, self._yc) )
            utils.brushExtFillPolygon(painter, extfill, cliprect, pts)
        elif filltype == 'outside':
            pp = qt.QPainterPath()
            pp.moveTo(self._xc, self._yc)
            pp.arcTo(cliprect, 0, 360)
            pp.addPolygon(pts)
            utils.brushExtFillPath(painter, extfill, pp)
        elif filltype == 'polygon':
            utils.brushExtFillPolygon(painter, extfill, cliprect, pts)

    def drawGraph(self, painter, bounds, datarange, outerbounds=None):
        '''Plot graph area and axes.'''

        s = self.settings

        if datarange is None:
            datarange = [0., 1., 0., 1.]

        if s.maxradius == 'Auto':
            self._maxradius = datarange[1]
        else:
            self._maxradius = s.maxradius

        if s.minradius == 'Auto':
            if s.log:
                if datarange[0] > 0.:
                    self._minradius = datarange[0]
                else:
                    self._minradius = self._maxradius / 100.
            else:
                if datarange[0] >= 0:
                    self._minradius = 0.
                else:
                    self._minradius = datarange[0]
        else:
            self._minradius = s.minradius

        # stop negative values
        if s.log:
            self._minradius = N.clip(self._minradius, 1e-99, 1e99)
            self._maxradius = N.clip(self._maxradius, 1e-99, 1e99)
        if self._minradius == self._maxradius:
            self._maxradius = self._minradius + 1

        self._xscale = (bounds[2]-bounds[0])*0.5
        self._yscale = (bounds[3]-bounds[1])*0.5
        self._xc = 0.5*(bounds[0]+bounds[2])
        self._yc = 0.5*(bounds[3]+bounds[1])

        path = qt.QPainterPath()
        path.addEllipse(qt.QRectF(
            qt.QPointF(bounds[0], bounds[1]),
            qt.QPointF(bounds[2], bounds[3]) ))
        utils.brushExtFillPath(
            painter, s.Background, path, stroke=s.Border.makeQPenWHide(painter))

    def setClip(self, painter, bounds):
        '''Set clipping for graph.'''
        p = qt.QPainterPath()
        p.addEllipse(qt.QRectF(
            qt.QPointF(bounds[0], bounds[1]),
            qt.QPointF(bounds[2], bounds[3]) ))
        painter.setClipPath(p)

    def drawAxes(self, painter, bounds, datarange, outerbounds=None):
        '''Plot axes.'''

        s = self.settings

        spokesL = s.SpokeLine
        radiiL = s.RadiiLine

        # handle reversed axes using min and max below
        r = [self._minradius, self._maxradius]
        atick = AxisTicks(
            min(r), max(r),
            radiiL.number, radiiL.number*4,
            extendmin=False, extendmax=False,
            logaxis=s.log)
        atick.getTicks()
        majtick = atick.tickvals

        # drop 0 at origin
        if self._minradius == 0. and not s.log:
            majtick = majtick[1:]

        # pen for radii circles and axis
        painter.setPen( radiiL.makeQPenWHide(painter) )
        painter.setBrush( qt.QBrush() )

        # draw ticks as circles
        if not radiiL.hide:
            for tick in majtick:
                radius = self.toPlotRadius(tick)
                if radius > 0:
                    rect = qt.QRectF(
                        qt.QPointF(
                            self._xc - radius*self._xscale,
                            self._yc - radius*self._yscale ),
                        qt.QPointF(
                            self._xc + radius*self._xscale,
                            self._yc + radius*self._yscale ) )
                    painter.drawEllipse(rect)

        # setup axes plot
        tl = s.TickLabels
        scale, fmt = tl.scale, tl.format
        if fmt == 'Auto':
            fmt = atick.autoformat
        painter.setPen( tl.makeQPen(painter) )
        font = tl.makeQFont(painter)

        # draw ticks
        if not s.TickLabels.hideradial:
            for tick in majtick:
                num = utils.formatNumber(
                    tick*scale, fmt,locale=self.document.locale)
                x = self.toPlotRadius(tick) * self._xscale + self._xc
                r = utils.Renderer(
                    painter, font, x, self._yc, num,
                    alignhorz=-1,
                    alignvert=-1, usefullheight=True,
                    doc=self.document)
                r.render()

        numspokes = spokesL.number
        if s.units == 'degrees':
            vals = N.linspace(0., 360., numspokes+1)[:-1]
            labels = ['%g°' % x for x in vals]
        elif s.units == 'radians':
            labels = []
            for i in range(numspokes):
                # use fraction module to work out labels
                # angle in radians (/pi)
                f = fractions.Fraction(2*i, numspokes)
                if f.numerator == 0:
                    txt = '0'
                else:
                    txt = '%iπ/%i' % (f.numerator, f.denominator)
                    # remove superfluous 1* or /1
                    if txt[-2:] == '/1': txt = txt[:-2]
                    txt = txt.lstrip('1')
                labels.append(txt)
        elif s.units == 'fractions':
            labels = []
            for i in range(numspokes):
                val = i/numspokes
                label = ('%.2f' % val).rstrip('0').rstrip('.')
                labels.append(label)
        elif s.units == 'percentages':
            labels = []
            for i in range(numspokes):
                txt = '%.1f' % (100*i/numspokes)
                if txt[-2:] == '.0': txt = txt[:-2]
                labels.append(txt)
        else:
            raise RuntimeError()

        if s.direction == 'anticlockwise':
            labels = labels[0:1] + labels[1:][::-1]

        # these are the angles the spokes lie in
        # (by default 0 is to the right)
        angles = 2 * math.pi * N.arange(numspokes) / numspokes

        # rotate labels if zero not at right
        if s.position0 == 'top':
            angles -= math.pi/2
        elif s.position0 == 'left':
            angles += math.pi
        elif s.position0 == 'bottom':
            angles += math.pi/2

        # draw labels around plot
        if not s.TickLabels.hidetangential:
            for angle, label in zip(angles, labels):
                align = radianToAlign(angle)
                x = self._xc +  N.cos(angle) * self._xscale
                y = self._yc +  N.sin(angle) * self._yscale
                r = utils.Renderer(
                    painter, font, x, y, label,
                    alignhorz=align[0],
                    alignvert=align[1],
                    usefullheight=True)
                r.render()

        # draw spokes
        if not spokesL.hide:
            painter.setPen( spokesL.makeQPenWHide(painter) )
            painter.setBrush( qt.QBrush() )
            angle = 2 * math.pi / numspokes
            lines = []
            for i in range(numspokes):
                x = self._xc +  N.cos(angle*i) * self._xscale
                y = self._yc +  N.sin(angle*i) * self._yscale
                lines.append(
                    qt.QLineF(
                        qt.QPointF(self._xc, self._yc), qt.QPointF(x, y)) )
            painter.drawLines(lines)

document.thefactory.register(Polar)



#    Copyright (C) 2008 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

"""For plotting shapes."""

import codecs
import itertools
import os

from .. import qtall as qt
from .. import setting
from .. import document
from .. import utils
from . import widget
from . import controlgraph
from . import plotters

def _(text, disambiguation=None, context='Shape'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class Shape(plotters.FreePlotter):
    """A shape on a page/graph."""

    def __init__(self, parent, name=None):
        plotters.FreePlotter.__init__(self, parent, name=name)

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        plotters.FreePlotter.addSettings(s)

        s.add( setting.ShapeFill(
            'Fill',
            descr = _('Shape fill'),
            usertext=_('Fill')),
            pixmap = 'settings_bgfill' )
        s.add( setting.Line(
            'Border',
            descr = _('Shape border'),
            usertext=_('Border')),
            pixmap = 'settings_border' )
        s.add( setting.Bool(
            'clip', False,
            descr=_('Clip shape to its container'),
            usertext=_('Clip'),
            formatting=True) )

class BoxShape(Shape):
    """For drawing box-like shapes."""

    def __init__(self, parent, name=None):
        Shape.__init__(self, parent, name=name)

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        Shape.addSettings(s)

        s.add( setting.DatasetExtended(
            'width', [0.1],
            descr=_('List of fractional widths, dataset or expression'),
            usertext=_('Widths'),
            formatting=False), 3 )
        s.add( setting.DatasetExtended(
            'height', [0.1],
            descr=_('List of fractional heights, dataset or expression'),
            usertext=_('Heights'),
            formatting=False), 4 )
        s.add( setting.DatasetExtended(
            'rotate', [0.],
            descr=_('Rotation angles of shape, dataset or expression'),
            usertext=_('Rotate'),
            formatting=False), 5 )

    def drawShape(self, painter, rect):
        pass

    def draw(self, posn, phelper, outerbounds = None):
        """Plot the key on a plotter."""

        s = self.settings
        d = self.document
        if s.hide:
            return

        # get positions of shapes
        width = s.get('width').getFloatArray(d)
        height = s.get('height').getFloatArray(d)
        rotate = s.get('rotate').getFloatArray(d)
        if width is None or height is None or rotate is None:
            return

        # translate coordinates from axes or relative values
        xpos, ypos = self._getPlotterCoords(posn)
        if xpos is None or ypos is None:
            # we can't calculate coordinates
            return

        # if a dataset is used, we can't use control items
        isnotdataset = (
            not s.get('xPos').isDataset(d) and
            not s.get('yPos').isDataset(d) and
            not s.get('width').isDataset(d) and
            not s.get('height').isDataset(d) and
            not s.get('rotate').isDataset(d)
        )

        clip = None
        if s.clip:
            clip = qt.QRectF(
                qt.QPointF(posn[0], posn[1]), qt.QPointF(posn[2], posn[3]))
        painter = phelper.painter(self, posn, clip=clip)
        with painter:
            # drawing settings for shape
            if not s.Border.hide:
                painter.setPen( s.get('Border').makeQPen(painter) )
            else:
                painter.setPen( qt.QPen(qt.Qt.NoPen) )

            # iterate over positions
            index = 0
            dx, dy = posn[2]-posn[0], posn[3]-posn[1]
            x = y = w = h = r = None
            for x, y, w, h, r in zip(
                    xpos, ypos,
                    itertools.cycle(width),
                    itertools.cycle(height),
                    itertools.cycle(rotate)):
                wp, hp = dx*w, dy*h
                painter.save()
                painter.translate(x, y)
                if r != 0:
                    painter.rotate(r)
                self.drawShape(painter, qt.QRectF(-wp*0.5, -hp*0.5, wp, hp))
                painter.restore()

        controlgraphitems = []
        if x is not None and isnotdataset:
            cgi = controlgraph.ControlResizableBox(
                self, phelper, [x, y], [wp, hp], r, allowrotate=True)
            cgi.index = index
            cgi.widgetposn = posn
            index += 1
            controlgraphitems.append(cgi)

        phelper.setControlGraph(self, controlgraphitems)

    def updateControlItem(self, cgi):
        """If control item is moved or resized, this is called."""
        s = self.settings

        # calculate new position coordinate for item
        xpos, ypos = self._getGraphCoords(
            cgi.widgetposn, cgi.posn[0], cgi.posn[1])
        if xpos is None or ypos is None:
            return

        xw = abs(cgi.dims[0] / (cgi.widgetposn[2]-cgi.widgetposn[0]))
        yw = abs(cgi.dims[1] / (cgi.widgetposn[1]-cgi.widgetposn[3]))

        # actually do the adjustment on the document
        xp = list(s.get('xPos').getFloatArray(self.document))
        yp = list(s.get('yPos').getFloatArray(self.document))
        w = list(s.get('width').getFloatArray(self.document))
        h = list(s.get('height').getFloatArray(self.document))
        r = list(s.get('rotate').getFloatArray(self.document))

        xp[min(cgi.index, len(xp)-1)] = xpos
        yp[min(cgi.index, len(yp)-1)] = ypos
        w[min(cgi.index, len(w)-1)] = xw
        h[min(cgi.index, len(h)-1)] = yw
        r[min(cgi.index, len(r)-1)] = cgi.angle

        operations = (
            document.OperationSettingSet(s.get('xPos'), xp),
            document.OperationSettingSet(s.get('yPos'), yp),
            document.OperationSettingSet(s.get('width'), w),
            document.OperationSettingSet(s.get('height'), h),
            document.OperationSettingSet(s.get('rotate'), r)
        )
        self.document.applyOperation(
            document.OperationMultiple(operations, descr=_('adjust shape')) )

class Rectangle(BoxShape):
    """Draw a rectangle, or rounded rectangle."""
    typename = 'rect'
    description = _('Rectangle')
    allowusercreation = True

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        BoxShape.addSettings(s)

        s.add( setting.Int(
            'rounding', 0,
            minval=0, maxval=100,
            descr=_('Round corners with this percentage'),
            usertext=_('Rounding corners'),
            formatting=True) )

    def drawShape(self, painter, rect):
        s = self.settings
        path = qt.QPainterPath()
        if s.rounding == 0:
            path.addRect(rect)
        else:
            path.addRoundedRect(rect, s.rounding, s.rounding)

        utils.brushExtFillPath(painter, s.Fill, path, stroke=painter.pen())

class Ellipse(BoxShape):
    """Draw an ellipse."""

    typename = 'ellipse'
    description = _('Ellipse')
    allowusercreation = True

    def drawShape(self, painter, rect):
        s = self.settings
        path = qt.QPainterPath()
        path.addEllipse(rect)
        utils.brushExtFillPath(painter, s.Fill, path, stroke=painter.pen())

class ImageFile(BoxShape):
    """Draw an image."""

    typename = 'imagefile'
    description = _('Image file')
    allowusercreation = True

    def __init__(self, parent, name=None):
        BoxShape.__init__(self, parent, name=name)

        self.cacheimage = None
        self.cachefilename = None
        self.cachestat = None
        self.cacheembeddata = None

        self.addAction( widget.Action(
            'embed', self.actionEmbed,
            descr=_(
                'Embed image in Veusz document '
                'to remove dependency on external file'),
            usertext=_('Embed image')) )

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        BoxShape.addSettings(s)

        s.add( setting.ImageFilename(
            'filename', '',
            descr=_('Image filename'),
            usertext=_('Filename'),
            formatting=False),
            posn=0 )

        s.add( setting.Str(
            'embeddedImageData', '',
            descr=_(
                'Embedded base 64-encoded image data, '
                'used if filename set to {embedded}'),
            usertext=_('Embedded data'),
            hidden=True) )

        s.add( setting.Bool(
            'aspect', True,
            descr=_('Preserve aspect ratio'),
            usertext=_('Preserve aspect'),
            formatting=True),
            posn=0 )
        s.Border.get('hide').newDefault(True)

    def actionEmbed(self):
        """Embed external image into veusz document."""

        s = self.settings

        if s.filename == '{embedded}':
            print("Data already embedded")
            return

        # get data from external file
        try:
            f = open(s.filename, 'rb')
            data = f.read()
            f.close()
        except EnvironmentError:
            print("Could not find file. Not embedding.")
            return

        # convert to base 64 to make it nicer in the saved file
        encoded = codecs.encode(data, 'base64').decode('ascii')

        # now put embedded data in hidden setting
        ops = [
            document.OperationSettingSet(s.get('filename'), '{embedded}'),
            document.OperationSettingSet(
                s.get('embeddedImageData'), encoded)
        ]
        self.document.applyOperation(
            document.OperationMultiple(ops, descr=_('embed image')) )

    def updateCachedImage(self):
        """Update cache."""
        s = self.settings
        self.cachestat = os.stat(s.filename)
        self.cacheimage = qt.QImage(s.filename)
        self.cachefilename = s.filename

    def updateCachedEmbedded(self):
        """Update cached image from embedded data."""
        s = self.settings
        self.cacheimage = qt.QImage()

        # convert the embedded data from base64 and load into the image
        decoded = codecs.decode(s.embeddedImageData.encode('ascii'), 'base64')
        self.cacheimage.loadFromData(decoded)

        # we cache the data we have decoded
        self.cacheembeddata = s.embeddedImageData

    def drawShape(self, painter, rect):
        """Draw image."""
        s = self.settings

        # draw border and fill
        painter.drawRect(rect)

        # check to see whether image needs reloading
        image = None
        if s.filename != '' and os.path.isfile(s.filename):
            if (self.cachefilename != s.filename or
                os.stat(s.filename) != self.cachestat):
                # update the image cache
                self.updateCachedImage()
                # clear any embedded image data
                self.settings.get('embeddedImageData').set('')
            image = self.cacheimage

        # or needs recreating from embedded data
        if s.filename == '{embedded}':
            if s.embeddedImageData is not self.cacheembeddata:
                self.updateCachedEmbedded()
            image = self.cacheimage

        # if no image, then use default image
        if ( not image or image.isNull() or
             image.width() == 0 or image.height() == 0 ):
            # load replacement image
            fname = os.path.join(utils.imagedir, 'button_imagefile.svg')
            r = qt.QSvgRenderer(fname)
            r.render(painter, rect)

        else:
            # image rectangle
            irect = qt.QRectF(image.rect())

            # preserve aspect ratio
            if s.aspect:
                xr = rect.width() / irect.width()
                yr = rect.height() / irect.height()

                if xr > yr:
                    rect = qt.QRectF(
                        rect.left()+(rect.width()-irect.width()*yr)*0.5,
                        rect.top(), irect.width()*yr, rect.height())
                else:
                    rect = qt.QRectF(
                        rect.left(),
                        rect.top()+(rect.height()-irect.height()*xr)*0.5,
                        rect.width(), irect.height()*xr)

            # finally draw image
            painter.drawImage(rect, image, irect)

class SVGFile(BoxShape):
    """Draw an scalable vector graphic."""

    typename = 'svgfile'
    description = _('Scalable vector graphic file')
    allowusercreation = True

    def __init__(self, parent, name=None):
        BoxShape.__init__(self, parent, name=name)

        self.cacheimage = None
        self.cachefilename = None
        self.cachestat = None
        self.cacheembeddata = None

        self.addAction( widget.Action(
            'embed', self.actionEmbed,
            descr=_(
                'Embed scalable vector graphic in Veusz document '
                'to remove dependency on external file'),
            usertext=_('Embed SVG')) )

    @classmethod
    def addSettings(klass, s):
        """Construct list of settings."""
        BoxShape.addSettings(s)

        s.add( setting.SVGFilename(
            'filename', '',
            descr=_('SVG filename'),
            usertext=_('Filename'),
            formatting=False),
            posn=0 )

        s.add( setting.Str(
            'embeddedSVGData', '',
            descr=_(
                'Embedded scalable vector graphic data, '
                'used if filename set to {embedded}'),
            usertext=_('Embedded data'),
            hidden=True) )

        s.add( setting.Bool(
            'aspect', True,
            descr=_('Preserve aspect ratio'),
            usertext=_('Preserve aspect'),
            formatting=True),
            posn=0 )
        s.Border.get('hide').newDefault(True)

    def actionEmbed(self):
        """Embed external vector graphic into veusz document."""

        s = self.settings

        if s.filename == '{embedded}':
            print("Data already embedded")
            return

        # get data from external file
        try:
            with open(s.filename, 'rb') as f:
                data = f.read()
        except EnvironmentError:
            print("Could not find file. Not embedding.")
            return

        # convert to base 64 to make it nicer in the saved file
        encoded = codecs.encode(data, 'base64').decode('ascii')

        # now put embedded data in hidden setting
        ops = [
            document.OperationSettingSet(s.get('filename'), '{embedded}'),
            document.OperationSettingSet(
                s.get('embeddedSVGData'), encoded)
        ]
        self.document.applyOperation(
            document.OperationMultiple(ops, descr=_('embed SVG')) )

    def updateCachedImage(self):
        """Update cache."""
        s = self.settings
        self.cachestat = os.stat(s.filename)
        self.cacheimage = qt.QSvgRenderer(s.filename)
        self.cachefilename = s.filename

    def updateCachedEmbedded(self):
        """Update cached image from embedded data."""
        s = self.settings
        self.cacheimage = qt.QSvgRenderer()

        # convert the embedded data from base64 and load into the SVG image
        binarized = qt.QByteArray.fromBase64(s.embeddedSVGData.encode('ascii'))
        self.cacheimage.load(binarized)

        # we cache the data we have decoded
        self.cacheembeddata = s.embeddedSVGData

    def drawShape(self, painter, rect):
        """Draw image."""
        s = self.settings

        # draw border and fill
        painter.drawRect(rect)

        # check to see whether image needs reloading
        image = None
        if s.filename != '' and os.path.isfile(s.filename):
            if (self.cachefilename != s.filename or
                os.stat(s.filename) != self.cachestat):
                # update the image cache
                self.updateCachedImage()
                # clear any embedded image data
                self.settings.get('embeddedSVGData').set('')
            image = self.cacheimage

        # or needs recreating from embedded data
        if s.filename == '{embedded}':
            if s.embeddedSVGData is not self.cacheembeddata:
                self.updateCachedEmbedded()
            image = self.cacheimage

        # if no image, then use default image
        if ( not image or not image.isValid() or
             image.viewBox().width() == 0 or image.viewBox().height() == 0 ):
            # load replacement image
            fname = os.path.join(utils.imagedir, 'button_svgfile.svg')
            r = qt.QSvgRenderer(fname)
            r.render(painter, rect)

        else:
            # image rectangle
            irect = qt.QRectF(image.viewBox())

            # preserve aspect ratio
            if s.aspect:
                xr = rect.width() / irect.width()
                yr = rect.height() / irect.height()

                if xr > yr:
                    rect = qt.QRectF(
                        rect.left()+(rect.width()-irect.width()*yr)*0.5,
                        rect.top(), irect.width()*yr, rect.height())
                else:
                    rect = qt.QRectF(
                        rect.left(),
                        rect.top()+(rect.height()-irect.height()*xr)*0.5,
                        rect.width(), irect.height()*xr)

            # finally draw image
            image.render(painter, rect)

document.thefactory.register(Ellipse)
document.thefactory.register(Rectangle)
document.thefactory.register(ImageFile)
document.thefactory.register(SVGFile)



#    Copyright (C) 2010 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Non orthogonal graph root."""

from . import controlgraph
from .widget import Widget

from .. import qtall as qt
from .. import setting

filloptions = (
    'center', 'outside', 'top', 'bottom', 'left', 'right',
    'polygon',
)

def _(text, disambiguation=None, context='NonOrthGraph'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

class FillBrush(setting.BrushExtended):
    '''Brush for filling point region.'''
    def __init__(self, *args, **argsv):
        setting.BrushExtended.__init__(self, *args, **argsv)
        self.add( setting.Choice(
            'filltype', filloptions, 'center',
            descr=_('Fill to this edge/position'),
            usertext=_('Fill type')) )
        self.get('hide').newDefault(True)

class NonOrthGraph(Widget):
    '''Non-orthogonal graph base widget.'''

    @classmethod
    def addSettings(klass, s):
        '''Construct list of settings.'''
        Widget.addSettings(s)

        s.add( setting.Distance(
            'leftMargin',
            '1.7cm',
            descr=_('Distance from left of graph to edge'),
            usertext=_('Left margin'),
            formatting=True) )
        s.add( setting.Distance(
            'rightMargin',
            '0.2cm',
            descr=_('Distance from right of graph to edge'),
            usertext=_('Right margin'),
            formatting=True) )
        s.add( setting.Distance(
            'topMargin',
            '0.2cm',
            descr=_('Distance from top of graph to edge'),
            usertext=_('Top margin'),
            formatting=True) )
        s.add( setting.Distance(
            'bottomMargin',
            '1.7cm',
            descr=_('Distance from bottom of graph to edge'),
            usertext=_('Bottom margin'),
            formatting=True) )
        s.add( setting.GraphBrush(
            'Background',
            descr=_('Background plot fill'),
            usertext=_('Background')),
            pixmap='settings_bgfill' )
        s.add( setting.Line(
            'Border', descr=_('Graph border line'),
            usertext=_('Border')),
            pixmap='settings_border')

    @classmethod
    def allowedParentTypes(klass):
        from . import page, grid
        return (page.Page, grid.Grid)

    def graphToPlotCoords(self, coorda, coordb):
        '''Convert graph to plotting coordinates.
        Returns (plta, pltb) coordinates
        '''

    def coordRanges(self):
        '''Return coordinate ranges of plot.
        This is in the form [[mina, maxa], [minb, maxb]].'''

    def drawFillPts(self, painter, extfill, bounds, ptsx, ptsy):
        '''Draw set of points for filling.
        extfill: extended fill brush
        bounds: usual tuple (minx, miny, maxx, maxy)
        ptsx, ptsy: translated plotter coordinates
        '''

    def drawGraph(self, painter, bounds, datarange, outerbounds=None):
        '''Plot graph area.
        datarange is  [mina, maxa, minb, maxb] or None
        '''

    def drawAxes(self, painter, bounds, datarange, outerbounds=None):
        '''Plot axes.
        datarange is  [mina, maxa, minb, maxb] or None
        '''

    def setClip(self, painter, bounds):
        '''Set clipping for graph.'''

    def getDataRange(self):
        """Get automatic data range. Return None if no data."""

        drange = [1e199, -1e199, 1e199, -1e199]
        for c in self.children:
            if hasattr(c, 'updateDataRanges'):
                c.updateDataRanges(drange)

        # no data
        if drange[0] > drange[1] or drange[2] > drange[3]:
            drange = None

        return drange

    def getMargins(self, painthelper):
        """Use settings to compute margins."""
        s = self.settings
        return (
            s.get('leftMargin').convert(painthelper),
            s.get('topMargin').convert(painthelper),
            s.get('rightMargin').convert(painthelper),
            s.get('bottomMargin').convert(painthelper)
        )

    def draw(self, parentposn, phelper, outerbounds=None):
        '''Update the margins before drawing.'''

        s = self.settings

        bounds = self.computeBounds(parentposn, phelper)
        maxbounds = self.computeBounds(parentposn, phelper, withmargin=False)

        # do no painting if hidden
        if s.hide:
            return bounds

        painter = phelper.painter(self, bounds)
        with painter:
            # reset counter and compute automatic colors
            phelper.autoplottercount = 0
            for c in self.children:
                c.setupAutoColor(painter)

            # plot graph
            datarange = self.getDataRange()
            self.drawGraph(painter, bounds, datarange, outerbounds=outerbounds)
            self.drawAxes(painter, bounds, datarange, outerbounds=outerbounds)

            # paint children
            for c in reversed(self.children):
                c.draw(bounds, phelper, outerbounds=outerbounds)

        # controls for adjusting margins
        phelper.setControlGraph(self, [
            controlgraph.ControlMarginBox(self, bounds, maxbounds, phelper)])

        return bounds

    def updateControlItem(self, cgi):
        """Graph resized or moved - call helper routine to move self."""
        cgi.setWidgetMargins()



# A module for embedding Veusz within another python program

#    Copyright (C) 2005 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""This module allows veusz to be embedded within other Python programs.
For example:

import time
import numpy
import veusz.embed as veusz

g = veusz.Embedded('new win')
g.To( g.Add('page') )
g.To( g.Add('graph') )
g.SetData('x', numpy.arange(20))
g.SetData('y', numpy.arange(20)**2)
g.Add('xy')
g.Zoom(0.5)

time.sleep(60)
g.Close()

More than one embedded window can be opened at once
"""

import atexit
import sys
import os
import os.path
import struct
import socket
import subprocess
import time
import uuid
import functools
import types
import pickle

# check remote process has this API version
API_VERSION = 2

def findOnPath(cmd):
    """Find a command on the system path, or None if does not exist."""
    path = os.getenv('PATH', os.path.defpath)
    pathparts = path.split(os.path.pathsep)
    for dirname in pathparts:
        dirname = dirname.strip('"')
        cmdtry = os.path.join(dirname, cmd)
        if os.path.isfile(cmdtry) and os.access(cmdtry, os.X_OK):
            return cmdtry
    return None

class Embedded:
    """An embedded instance of Veusz.

    This embedded instance supports all the normal veusz functions
    """

    remote = None

    def __init__(self, name='Veusz', copyof=None, hidden=False, compatlevel=0):
        """Initialse the embedded veusz window.

        name is the name of the window to show.
        copyof duplicates a view of the document in the Embedded instance given
        hidden makes a hidden window (useful for batch scripting)
        compatlevel is the compatibility level to use (-1 is latest)
        """

        if not Embedded.remote:
            Embedded.startRemote()

        if not copyof:
            retval = self.sendCommand(
                (-1, '_NewWindow', (name,), {
                    'hidden': hidden,
                }) )
        else:
            retval = self.sendCommand(
                (-1, '_NewWindowCopy', (name, copyof.winno), {
                    'hidden': hidden,
                }) )

        self.winno, cmds = retval

        # add methods corresponding to Veusz commands
        for name, doc in cmds:
            func = functools.partial(self.runCommand, name)
            func.__doc__ = doc    # set docstring
            func.__name__ = name  # make name match what it calls
            method =  types.MethodType(func, self)
            setattr(self, name, method) # assign to self

        # check API version is same
        try:
            remotever = self._apiVersion()
        except AttributeError:
            remotever = 0
        if remotever != API_VERSION:
            raise RuntimeError(
                "Remote Veusz instance reports version %i of"
                " API. This embed.py supports version %i." %
                (remotever, API_VERSION)
            )

        # increase compatibility level (if requested)
        if compatlevel != 0:
            self.SetCompatLevel(compatlevel)

        # define root object
        self.Root = WidgetNode(self, 'widget', '/')

    def StartSecondView(self, name = 'Veusz'):
        """Provides a second view onto the document of this window.

        Returns an Embedded instance
        """
        return Embedded(name=name, copyof=self)

    def WaitForClose(self):
        """Wait for the window to close."""

        # this is messy, polling for closure, but cleaner than doing
        # it in the remote client
        while not self.IsClosed():
            time.sleep(0.1)

    @classmethod
    def makeSockets(cls):
        """Make socket(s) to communicate with remote process.
        Returns string to send to remote process
        """

        if ( hasattr(socket, 'AF_UNIX') and hasattr(socket, 'socketpair') ):
            # convenient interface
            cls.sockfamily = socket.AF_UNIX
            sock, socket2 = socket.socketpair(cls.sockfamily, socket.SOCK_STREAM)

            # socket is closed on popen in Python 3.4+ without this (PEP 446)
            try:
                os.set_inheritable(socket2.fileno(), True)
            except AttributeError:
                pass

            sendtext = 'unix %i\n' % socket2.fileno()
            cls.socket2 = socket2    # prevent socket being destroyed
            waitaccept = False

        else:
            # otherwise mess around with internet sockets
            # * This is required for windows, which doesn't have AF_UNIX
            # * It is required where socketpair is not supported
            cls.sockfamily = socket.AF_INET
            sock = socket.socket(cls.sockfamily, socket.SOCK_STREAM)
            sock.bind( ('localhost', 0) )
            interface, port = sock.getsockname()
            sock.listen(1)
            sendtext = 'internet %s %i\n' % (interface, port)
            waitaccept = True

        return (sock, sendtext.encode('ascii'), waitaccept)

    @classmethod
    def makeRemoteProcess(cls):
        """Try to find veusz process for remote program."""

        # here's where to look for embed_remote.py
        thisdir = os.path.dirname(os.path.abspath(__file__))

        # build up a list of possible command lines to start the remote veusz
        if sys.platform == 'win32':
            # windows is a special case
            # we need to run embed_remote.py under pythonw.exe, not python.exe

            # look for the python windows interpreter on path
            findpython = findOnPath('pythonw.exe')
            if not findpython:
                # if it wasn't on the path, use sys.prefix instead
                findpython = os.path.join(sys.prefix, 'pythonw.exe')

            # look for veusz executable on path
            findexe = findOnPath('veusz.exe')
            if not findexe:
                try:
                    # add the usual place as a guess :-(
                    findexe = os.path.join(
                        os.environ['ProgramFiles'],
                        'Veusz', 'veusz.exe')
                except KeyError:
                    pass

            # here is the list of commands to try
            possiblecommands = [
                [ findpython, os.path.join(thisdir, 'veusz_main.py') ],
                [ findexe ],
            ]

        else:
            executable = sys.executable

            # try embed_remote.py in this directory, veusz in this directory
            # or veusz on the path in order
            possiblecommands = [
                [ executable, os.path.join(thisdir, 'veusz_main.py') ],
                [ os.path.join(thisdir, 'veusz') ],
                [ findOnPath('veusz') ],
            ]

        # cheat and look for Veusz app for MacOS under the standard application
        # directory. I don't know how else to find it :-(
        if sys.platform == 'darwin':
            findbundle = findOnPath('Veusz.app')
            if findbundle:
                possiblecommands += [ [findbundle+'/Contents/MacOS/Veusz'] ]
            else:
                possiblecommands += [[
                    '/Applications/Veusz.app/Contents/MacOS/Veusz' ]]
                possiblecommands += [[
                    os.path.expanduser('~/Applications/Veusz.app/Contents/MacOS/Veusz')]]


        for cmd in possiblecommands:
            # only try to run commands that exist as error handling
            # does not work well when interfacing with OS (especially Windows)
            if ( None not in cmd and
                 False not in [os.path.isfile(c) for c in cmd] ):
                try:
                    # we don't use stdout below, but works around windows bug
                    # http://bugs.python.org/issue1124861
                    cls.remote = subprocess.Popen(
                        cmd + ['--embed-remote'],
                        shell=False, bufsize=0,
                        close_fds=False,
                        stdin=subprocess.PIPE,
                        stdout=subprocess.PIPE,
                    )
                    return
                except OSError:
                    pass

        raise RuntimeError('Unable to find a veusz executable on system path')

    @classmethod
    def startRemote(cls):
        """Start remote process."""
        cls.serv_socket, sendtext, waitaccept = cls.makeSockets()

        cls.makeRemoteProcess()
        stdin = cls.remote.stdin

        # send socket number over pipe
        stdin.write( sendtext )

        # accept connection if necessary
        if waitaccept:
            cls.serv_socket, address = cls.serv_socket.accept()

        # Send a secret to the remote program by secure route and
        # check it comes back.  This is to check that no program has
        # secretly connected on our port, which isn't really useful
        # for AF_UNIX sockets.
        secret = (str(uuid.uuid4()) + '\n').encode('ascii')
        stdin.write(secret)
        secretback = cls.readLenFromSocket(cls.serv_socket, len(secret))
        if secret != secretback:
            raise RuntimeError("Security between client and server broken")

        # packet length for command bytes
        cls.cmdlen = struct.calcsize('<I')
        atexit.register(cls.exitQt)

    @staticmethod
    def readLenFromSocket(socket, length):
        """Read length bytes from socket."""
        s = b''
        while len(s) < length:
            s += socket.recv(length-len(s))
        return s

    @staticmethod
    def writeToSocket(socket, data):
        count = 0
        while count < len(data):
            count += socket.send(data[count:])

    @classmethod
    def sendCommand(cls, cmd):
        """Send the command to the remote process."""

        # note: protocol 2 for python2 compat
        outs = pickle.dumps(cmd, 2)

        cls.writeToSocket( cls.serv_socket, struct.pack('<I', len(outs)) )
        cls.writeToSocket( cls.serv_socket, outs )

        backlen = struct.unpack('<I', cls.readLenFromSocket(
            cls.serv_socket, cls.cmdlen))[0]
        rets = cls.readLenFromSocket( cls.serv_socket, backlen )
        retobj = pickle.loads(rets)

        if isinstance(retobj, Exception):
            raise retobj
        else:
            return retobj

    def runCommand(self, cmd, *args, **args2):
        """Execute the given function in the Qt thread with the arguments
        given."""
        return self.sendCommand( (self.winno, cmd, args[1:], args2) )

    @classmethod
    def exitQt(cls):
        """Exit the Qt thread."""
        try:
            cls.sendCommand( (-1, '_Quit', (), {}) )
            cls.serv_socket.shutdown(socket.SHUT_RDWR)
            cls.serv_socket.close()
        except socket.error:
            pass
        cls.serv_socket, cls.from_pipe = -1, -1

############################################################################
# Tree-based interface to Veusz widget tree below

class Node:
    """Represents an element in the Veusz widget-settinggroup-setting tree."""

    def __init__(self, ci, wtype, path):
        self._ci = ci
        self._type = wtype
        self._path = path

    def __eq__(self, other):
        """Is this the same node?"""
        return self._ci is other._ci and self._path == other._path

    def __ne__(self, other):
        """Is this a different node?"""
        return self._ci is not other._ci or self._path != other._path

    def __repr__(self):
        """Text description"""
        return "<%s at %s (type %s)>" % (
            self.__class__.__name__, repr(self._path), self._type)

    def fromPath(self, path):
        """Return a new Node for the path given."""
        wtype = self._ci.NodeType(path)
        if wtype == 'widget':
            return WidgetNode(self._ci, wtype, path)
        elif wtype == 'setting':
            return SettingNode(self._ci, wtype, path)
        else:
            return SettingGroupNode(self._ci, wtype, path)

    @property
    def path(self):
        """Veusz full path to node"""
        return self._path

    @property
    def type(self):
        """Type of node: 'widget', 'settinggroup', or 'setting'"""
        return self._type

    def _joinPath(self, child):
        """Return new path of child."""
        if self._path == '/':
            return '/' + child
        else:
            return self._path + '/' + child

    def __getitem__(self, key):
        """Return a child widget, settinggroup or setting."""

        if self._type != 'setting':
            try:
                return self.fromPath(self._joinPath(key))
            except ValueError:
                pass

        raise KeyError("%s does not have key or child '%s'" % (
            self.__class__.__name__, key))

    def __getattr__(self, attr):
        """Return a child widget, settinggroup or setting."""

        if self._type == 'setting':
            pass
        elif attr[:2] != '__':
            try:
                return self.fromPath(self._joinPath(attr))
            except ValueError:
                pass

        raise AttributeError("%s does not have attribute or child '%s'" % (
            self.__class__.__name__, attr))

    # boring ways to get children of nodes
    @property
    def children(self):
        """Generator to get children as Nodes."""
        for c in self._ci.NodeChildren(self._path):
            yield self.fromPath(self._joinPath(c))
    @property
    def children_widgets(self):
        """Generator to get child widgets as Nodes."""
        for c in self._ci.NodeChildren(self._path, types='widget'):
            yield self.fromPath(self._joinPath(c))
    @property
    def children_settings(self):
        """Generator to get child settings as Nodes."""
        for c in self._ci.NodeChildren(self._path, types='setting'):
            yield self.fromPath(self._joinPath(c))
    @property
    def children_settinggroups(self):
        """Generator to get child settingsgroups as Nodes."""
        for c in self._ci.NodeChildren(self._path, types='settinggroup'):
            yield self.fromPath(self._joinPath(c))

    @property
    def childnames(self):
        """Get names of children."""
        return self._ci.NodeChildren(self._path)
    @property
    def childnames_widgets(self):
        """Get names of children widgets."""
        return self._ci.NodeChildren(self._path, types='widget')
    @property
    def childnames_settings(self):
        """Get names of child settings."""
        return self._ci.NodeChildren(self._path, types='setting')
    @property
    def childnames_settinggroups(self):
        """Get names of child setting groups"""
        return self._ci.NodeChildren(self._path, types='settinggroup')

    @property
    def parent(self):
        """Return parent of node."""
        if self._path == '/':
            raise TypeError("Cannot get parent node of root node""")
        p = self._path.split('/')[:-1]
        if p == ['']:
            newpath = '/'
        else:
            newpath = '/'.join(p)
        return self.fromPath(newpath)

    @property
    def name(self):
        """Get name of node."""
        if self._path == '/':
            return self._path
        else:
            return self._path.split('/')[-1]

class SettingNode(Node):
    """A node which is a setting."""

    def _getVal(self):
        """The value of a setting."""
        if self._type == 'setting':
            return self._ci.Get(self._path)
        raise TypeError("Cannot get value unless is a setting""")

    def _setVal(self, val):
        if self._type == 'setting':
            self._ci.Set(self._path, val)
        else:
            raise TypeError("Cannot set value unless is a setting.""")

    val = property(_getVal, _setVal)

    @property
    def isreference(self):
        """Is this setting set to a reference to another setting?."""
        ref = self._ci.ResolveReference(self._path)
        return bool(ref)

    def resolveReference(self):
        """If this is set to a reference to a setting, return a new
        SettingNode to the original setting.

        If there are a chain of references, follow them to the target.
        Returns None if this setting is not set to a reference.
        """

        real = self._ci.ResolveReference(self._path)
        if not real:
            return None
        return self.fromPath(real)

    def setToReference(self, othernode):
        """Make this setting point to another setting, by creating a
        reference.

        References can be chained. Note that the absolute path is used
        to specify a reference, so moving affected widgets around will
        destroy the link."""

        if not isinstance(othernode, SettingNode):
            raise ValueError("othernode is not a SettingNode")

        self._ci.SetToReference(self._path, othernode._path)

    @property
    def settingtype(self):
        """Get the type of setting, which is a string."""
        return self._ci.SettingType(self._path)

class SettingGroupNode(Node):
    """A node containing a group of settings."""

    pass

class WidgetNode(Node):
    """A node pointing to a widget."""

    @property
    def widgettype(self):
        """Get Veusz type of widget."""
        return self._ci.WidgetType(self.path)

    def WalkWidgets(self, widgettype=None):
        """Generator to walk widget tree and get this widget and the
        widgets below this WidgetNode of type given.

        widgettype is a Veusz widget type name or None to get all
        widgets."""

        if widgettype is None or self._ci.WidgetType(self._path) == widgettype:
            yield self
        for child in self.children_widgets:
            for w in child.WalkWidgets(widgettype=widgettype):
                yield w

    def Add(self, widgettype, *args, **args_opt):
        """Add a widget of the type given, returning the Node instance.
        """

        args_opt['widget'] = self._path
        name = self._ci.Add(widgettype, *args, **args_opt)
        return WidgetNode( self._ci, 'widget', self._joinPath(name) )

    def Rename(self, newname):
        """Renames widget to name given."""

        if self._path == '/':
            raise RuntimeError("Cannot rename root widget")

        self._ci.Rename(self._path, newname)
        self._path = '/'.join( self._path.split('/')[:-1] + [newname] )

    def Action(self, action):
        """Applies action on widget."""
        self._ci.Action(action, widget=self._path)

    def Remove(self):
        """Removes a widget and its children."""
        self._ci.Remove(self._path)

    def Clone(self, newparent, newname=None):
        """Clone widget, placing at newparent. Uses newname if given.

        Returns new node."""
        path = self._ci.CloneWidget(
            self._path, newparent._path, newname=newname)
        return WidgetNode( self._ci, 'widget', path )



#    Copyright (C) 2010 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Define a base dialog class cleans up self after being hidden."""

import os.path

from .. import qtall as qt
from .. import utils

# register functions to open up dialogs to recreate a dataset
recreate_register = {}

class VeuszDialog(qt.QDialog):
    """Base dialog class.
    - Loads self from ui file.
    - Deletes self on closing.
    - Emits dialogFinished when dialog is done
    """

    dialogFinished = qt.pyqtSignal(qt.QDialog)

    def __init__(self, mainwindow, uifile, modal=False):
        """Initialise dialog given Veusz mainwindow and uifile for dialog.
        If modal is False, base on a top level window instead
        """

        flag = qt.Qt.Dialog
        if not modal:
            flag |= (
                qt.Qt.CustomizeWindowHint | qt.Qt.WindowMinimizeButtonHint |
                qt.Qt.WindowMaximizeButtonHint | qt.Qt.WindowCloseButtonHint |
                qt.Qt.WindowTitleHint | qt.Qt.WindowSystemMenuHint
            )

        qt.QDialog.__init__(self, mainwindow, flag)
        self.setAttribute(qt.Qt.WA_DeleteOnClose)

        qt.loadUi(os.path.join(utils.resourceDirectory, 'ui', uifile), self)

        self.mainwindow = mainwindow

    def hideEvent(self, event):
        """Emits dialogFinished if hidden."""
        if not event.spontaneous():
            self.dialogFinished.emit(self)
        return qt.QDialog.hideEvent(self, event)



#    Copyright (C) 2003 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""Implements the main window of the application."""

import os
import os.path
import sys
import re
import datetime

try:
    import h5py
except ImportError:
    h5py = None

from .. import qtall as qt

from .. import document
from .. import utils
from ..utils import vzdbus
from .. import setting
from .. import plugins
from ..qtwidgets.clicklabel import ClickLabel

from . import consolewindow
from . import plotwindow
from . import treeeditwindow
from .datanavigator import DataNavigatorWindow

def _(text, disambiguation=None, context='MainWindow'):
    """Translate text."""
    return qt.QCoreApplication.translate(context, text, disambiguation)

# shortcut to this
setdb = setting.settingdb

class DBusWinInterface(vzdbus.Object):
    """Simple DBus interface to window for triggering actions."""

    interface = 'org.veusz.actions'

    def __init__(self, actions, index):
        prefix = '/Windows/%i/Actions' % index
        # possible exception in dbus means we have to check sessionbus
        if vzdbus.sessionbus is not None:
            vzdbus.Object.__init__(self, vzdbus.sessionbus, prefix)
        self.actions = actions

    @vzdbus.method(dbus_interface=interface, out_signature='as')
    def GetActions(self):
        """Get list of actions which can be activated."""
        return sorted(self.actions)

    @vzdbus.method(dbus_interface=interface, in_signature='s')
    def TriggerAction(self, action):
        """Activate action given."""
        self.actions[action].trigger()

class MainWindow(qt.QMainWindow):
    """ The main window class for the application."""

    # this is emitted when a dialog is opened by the main window
    dialogShown = qt.pyqtSignal(qt.QWidget)
    # emitted when a document is opened
    documentOpened = qt.pyqtSignal()

    windows = []
    @classmethod
    def CreateWindow(cls, filename=None, mode='graph'):
        """Window factory function.

        If filename is given then that file is loaded into the window.
        Returns window created
        """

        # create the window, and optionally load a saved file
        win = cls()
        win.show()
        if filename:
            # load document
            win.openFileInWindow(filename)
        else:
            win.setupDefaultDoc(mode)

        # try to select first graph of first page
        win.treeedit.doInitialWidgetSelect()

        cls.windows.append(win)

        # check if tutorial wanted (only for graph mode)
        if not setting.settingdb['ask_tutorial'] and mode=='graph':
            win.askTutorial()
            # don't ask again
            setting.settingdb['ask_tutorial'] = True

        # check if version check is ok
        win.askVersionCheck()
        # periodically do the check
        win.doVersionCheck()

        # is it ok to do feedback?
        win.askFeedbackCheck()
        # periodically send feedback
        win.doFeedback()

        return win

    def __init__(self, *args):
        qt.QMainWindow.__init__(self, *args)
        self.setAcceptDrops(True)

        # icon and different size variations
        self.setWindowIcon( utils.getIcon('veusz') )

        # master document
        self.document = document.Document()

        # filename for document and update titlebar
        self.filename = ''
        self.updateTitlebar()

        # keep a list of references to dialogs
        self.dialogs = []

        # construct menus and toolbars
        self._defineMenus()

        # make plot window
        self.plot = plotwindow.PlotWindow(
            self.document, self, menu=self.menus['view'])
        self.setCentralWidget(self.plot)
        self.plot.showToolbar()

        # likewise with the tree-editing window
        self.treeedit = treeeditwindow.TreeEditDock(self.document, self)
        self.addDockWidget(qt.Qt.LeftDockWidgetArea, self.treeedit)
        self.propdock = treeeditwindow.PropertiesDock(
            self.document, self.treeedit, self)
        self.addDockWidget(qt.Qt.LeftDockWidgetArea, self.propdock)
        self.formatdock = treeeditwindow.FormatDock(
            self.document, self.treeedit, self)
        self.addDockWidget(qt.Qt.LeftDockWidgetArea, self.formatdock)
        self.datadock = DataNavigatorWindow(self.document, self, self)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self.datadock)

        # make the console window a dock
        self.console = consolewindow.ConsoleWindow(
            self.document, self)
        self.console.hide()
        self.interpreter = self.console.interpreter
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self.console)

        # assemble the statusbar
        statusbar = self.statusbar = qt.QStatusBar(self)
        self.setStatusBar(statusbar)
        self.updateStatusbar(_('Ready'))

        # a label for the picker readout
        self.pickerlabel = qt.QLabel(statusbar)
        self._setPickerFont(self.pickerlabel)
        statusbar.addPermanentWidget(self.pickerlabel)
        self.pickerlabel.hide()

        # plot queue - how many plots are currently being drawn
        self.plotqueuecount = 0
        self.plot.sigQueueChange.connect(self.plotQueueChanged)
        self.plotqueuelabel = qt.QLabel()
        self.plotqueuelabel.setToolTip(_("Number of rendering jobs remaining"))
        statusbar.addWidget(self.plotqueuelabel)
        self.plotqueuelabel.show()

        # a label for the cursor position readout
        self.axisvalueslabel = qt.QLabel(statusbar)
        statusbar.addPermanentWidget(self.axisvalueslabel)
        self.axisvalueslabel.show()
        self.slotUpdateAxisValues(None)

        # a label for the page number readout
        self.pagelabel = qt.QLabel(statusbar)
        statusbar.addPermanentWidget(self.pagelabel)
        self.pagelabel.show()

        # security label
        self.securitylabel = ClickLabel(_("Untrusted mode"), statusbar)
        statusbar.addPermanentWidget(self.securitylabel)
        self.securitylabel.show()
        self.document.sigSecuritySet.connect(self.slotUpdateSecurity)
        self.securitylabel.clicked.connect(self.slotFileTrust)

        # working directory - use previous one
        self.dirname = setdb.get('dirname', qt.QDir.homePath())
        if setdb['dirname_usecwd']:
            self.dirname = os.getcwd()

        # connect plot signals to main window
        self.plot.sigUpdatePage.connect(self.slotUpdatePage)
        self.plot.sigAxisValuesFromMouse.connect(self.slotUpdateAxisValues)
        self.plot.sigPickerEnabled.connect(self.slotPickerEnabled)
        self.plot.sigPointPicked.connect(self.slotUpdatePickerLabel)

        # disable save if already saved
        self.document.signalModified.connect(self.slotModifiedDoc)
        # if the treeeditwindow changes the page, change the plot window
        self.treeedit.sigPageChanged.connect(self.plot.setPageNumber)

        # if a widget in the plot window is clicked by the user
        self.plot.sigWidgetClicked.connect(self.treeedit.selectWidget)
        self.treeedit.widgetsSelected.connect(self.plot.selectedWidgets)

        # enable/disable undo/redo
        self.menus['edit'].aboutToShow.connect(self.slotAboutToShowEdit)

        # get the list of recently opened files
        self.populateRecentFiles()
        self.setupWindowGeometry()
        self.defineViewWindowMenu()

        # if document requests it, ask whether an allowed import
        self.document.sigAllowedImports.connect(self.slotAllowedImportsDoc)

        # add on dbus interface
        self.dbusdocinterface = document.DBusInterface(self.document)
        self.dbuswininterface = DBusWinInterface(
            self.vzactions, self.dbusdocinterface.index)

        # has the document already been setup
        self.documentsetup = False

    def updateStatusbar(self, text):
        '''Display text for a set period.'''
        self.statusBar().showMessage(text, 2000)

    def dragEnterEvent(self, event):
        """Check whether event is valid to be dropped."""
        if event.mimeData().hasUrls() and self._getVeuszDropFiles(event):
            event.acceptProposedAction()

    def dropEvent(self, event):
        """Respond to a drop event on the current window"""
        if event.mimeData().hasUrls():
            files = self._getVeuszDropFiles(event)
            if files:
                if self.document.isBlank():
                    self.openFileInWindow(files[0])
                else:
                    self.CreateWindow(files[0])
                for filename in files[1:]:
                    self.CreateWindow(filename)

    def _getVeuszDropFiles(self, event):
        """Return a list of veusz files from a drag/drop event containing a
        text/uri-list"""

        mime = event.mimeData()
        if not mime.hasUrls():
            return []
        else:
            # get list of vsz files dropped
            urls = [u.toLocalFile() for u in mime.urls()]
            urls = [u for u in urls if os.path.splitext(u)[1] == '.vsz']
            return urls

    def setupDefaultDoc(self, mode):
        """Setup default document."""

        if not self.documentsetup:
            # add page and default graph
            self.document.makeDefaultDoc(mode)

            # set color theme
            self.document.basewidget.settings.get(
                'colorTheme').set(setting.settingdb['colortheme_default'])

            # load defaults if set
            self.loadDefaultStylesheet()
            self.loadDefaultCustomDefinitions()

            # done setup
            self.documentsetup = True

    def loadDefaultStylesheet(self):
        """Loads the default stylesheet for the new document."""
        filename = setdb['stylesheet_default']
        if filename:
            try:
                self.document.applyOperation(
                    document.OperationLoadStyleSheet(filename) )
            except EnvironmentError as e:
                qt.QMessageBox.warning(
                    self, _("Error - Veusz"),
                    _("Unable to load default stylesheet '%s'\n\n%s") %
                    (filename, e.strerror))
            else:
                # reset any modified flag
                self.document.setModified(False)
                self.document.changeset = 0

    def loadDefaultCustomDefinitions(self):
        """Loads the custom definitions for the new document."""
        filename = setdb['custom_default']
        if filename:
            try:
                self.document.applyOperation(
                    document.OperationLoadCustom(filename) )
            except EnvironmentError as e:
                qt.QMessageBox.warning(
                    self, _("Error - Veusz"),
                    _("Unable to load custom definitions '%s'\n\n%s") %
                    (filename, e.strerror))
            else:
                # reset any modified flag
                self.document.setModified(False)
                self.document.changeset = 0

    def slotAboutToShowEdit(self):
        """Enable/disable undo/redo menu items."""

        # enable distable, and add appropriate text to describe
        # the operation being undone/redone
        canundo = self.document.canUndo()
        undotext = _('Undo')
        if canundo:
            undotext = "%s %s" % (undotext, self.document.historyundo[-1].descr)
        self.vzactions['edit.undo'].setText(undotext)
        self.vzactions['edit.undo'].setEnabled(canundo)

        canredo = self.document.canRedo()
        redotext = _('Redo')
        if canredo:
            redotext = "%s %s" % (redotext, self.document.historyredo[-1].descr)
        self.vzactions['edit.redo'].setText(redotext)
        self.vzactions['edit.redo'].setEnabled(canredo)

    def slotEditUndo(self):
        """Undo the previous operation"""
        if self.document.canUndo():
            self.document.undoOperation()
        self.treeedit.checkWidgetSelected()

    def slotEditRedo(self):
        """Redo the previous operation"""
        if self.document.canRedo():
            self.document.redoOperation()

    def slotEditPreferences(self):
        from ..dialogs.preferences import PreferencesDialog
        dialog = PreferencesDialog(self)
        dialog.exec_()

    def slotEditStylesheet(self):
        from ..dialogs.stylesheet import StylesheetDialog
        dialog = StylesheetDialog(self, self.document)
        self.showDialog(dialog)
        return dialog

    def slotEditCustom(self):
        from ..dialogs.custom import CustomDialog
        dialog = CustomDialog(self, self.document)
        self.showDialog(dialog)
        return dialog

    def definePlugins(self, pluginlist, actions, menuname):
        """Create menu items and actions for plugins.

        pluginlist: list of plugin classes
        actions: dict of actions to add new actions to
        menuname: string giving prefix for new menu entries (inside actions)
        """

        def getLoadDialog(pluginkls):
            def _loadPlugin():
                from ..dialogs.plugin import handlePlugin
                handlePlugin(self, self.document, pluginkls)
            return _loadPlugin

        menu = []
        for pluginkls in pluginlist:
            actname = menuname + '.' + '.'.join(pluginkls.menu)
            text = pluginkls.menu[-1]
            if pluginkls.has_parameters:
                text += '...'
            actions[actname] = utils.makeAction(
                self,
                pluginkls.description_short,
                text,
                getLoadDialog(pluginkls))

            # build up menu from tuple of names
            menulook = menu
            namebuild = [menuname]
            for cmpt in pluginkls.menu[:-1]:
                namebuild.append(cmpt)
                name = '.'.join(namebuild)

                for c in menulook:
                    if c[0] == name:
                        menulook = c[2]
                        break
                else:
                    menulook.append( [name, cmpt, []] )
                    menulook = menulook[-1][2]

            menulook.append(actname)

        return menu

    def _defineMenus(self):
        """Initialise the menus and toolbar."""

        # these are actions for main menu toolbars and menus
        a = utils.makeAction
        self.vzactions = {
            'file.new.menu':
                a(self, _('New document'), _('New'),
                  None,
                  icon='kde-document-new'),
            'file.new.graph':
                a(self,
                  _('New graph document'),
                  _('&New graph document'),
                  self.slotFileNewGraph,
                  icon='kde-document-new-graph', key='Ctrl+N'),
            'file.new.polar':
                a(self,
                  _('New polar plot document'),
                  _('New polar document'),
                  self.slotFileNewPolar,
                  icon='kde-document-new-polar'),
            'file.new.ternary':
                a(self,
                  _('New ternary plot document'),
                  _('New ternary document'),
                  self.slotFileNewTernary,
                  icon='kde-document-new-ternary'),
            'file.new.graph3d':
                a(self,
                  _('New 3D plot document'),
                  _('New 3D document'),
                  self.slotFileNewGraph3D,
                  icon='kde-document-new-graph3d'),

            'file.open':
                a(self, _('Open a document'), _('&Open...'),
                  self.slotFileOpen,
                  icon='kde-document-open', key='Ctrl+O'),
            'file.reload':
                a(self, _('Reload document from saved version'),
                  _('Reload...'), self.slotFileReload),
            'file.save':
                a(self, _('Save the document'), _('&Save'),
                  self.slotFileSave,
                  icon='kde-document-save', key='Ctrl+S'),
            'file.saveas':
                a(self, _('Save the current document under a new name'),
                  _('Save &As...'), self.slotFileSaveAs,
                  icon='kde-document-save-as'),
            'file.trust':
                a(self, _('Trust document contents'), _('Trust...'),
                  self.slotFileTrust),
            'file.print':
                a(self, _('Print the document'), _('&Print...'),
                  self.slotFilePrint,
                  icon='kde-document-print', key='Ctrl+P'),
            'file.export':
                a(self, _('Export to graphics formats'), _('&Export...'),
                  self.slotFileExport,
                  icon='kde-document-export'),
            'file.close':
                a(self, _('Close current window'), _('Close Window'),
                  self.slotFileClose,
                  icon='kde-window-close', key='Ctrl+W'),
            'file.quit':
                a(self, _('Exit the program'), _('&Quit'),
                  self.slotFileQuit,
                  icon='kde-application-exit', key='Ctrl+Q'),

            'edit.undo':
                a(self, _('Undo the previous operation'), _('Undo'),
                  self.slotEditUndo,
                  icon='kde-edit-undo',  key='Ctrl+Z'),
            'edit.redo':
                a(self, _('Redo the previous operation'), _('Redo'),
                  self.slotEditRedo,
                  icon='kde-edit-redo', key='Ctrl+Shift+Z'),
            'edit.prefs':
                a(self, _('Edit preferences'), _('Preferences...'),
                  self.slotEditPreferences,
                  icon='veusz-edit-prefs'),
            'edit.custom':
                a(self,
                  _('Edit custom functions, constants, colors and colormaps'),
                  _('Custom definitions...'),
                  self.slotEditCustom,
                  icon='veusz-edit-custom'),

            'edit.stylesheet':
                a(self,
                  _('Edit stylesheet to change default widget settings'),
                  _('Default styles...'),
                  self.slotEditStylesheet, icon='settings_stylesheet'),

            'view.edit':
                a(self, _('Show or hide edit window'), _('Edit window'),
                  None, checkable=True),
            'view.props':
                a(self, _('Show or hide property window'), _('Properties window'),
                  None, checkable=True),
            'view.format':
                a(self, _('Show or hide formatting window'), _('Formatting window'),
                  None, checkable=True),
            'view.console':
                a(self, _('Show or hide console window'), _('Console window'),
                  None, checkable=True),
            'view.datanav':
                a(self, _('Show or hide data navigator window'), _('Data navigator window'),
                  None, checkable=True),

            'view.maintool':
                a(self, _('Show or hide main toolbar'), _('Main toolbar'),
                  None, checkable=True),
            'view.datatool':
                a(self, _('Show or hide data toolbar'), _('Data toolbar'),
                  None, checkable=True),
            'view.viewtool':
                a(self, _('Show or hide view toolbar'), _('View toolbar'),
                  None, checkable=True),
            'view.edittool':
                a(self, _('Show or hide editing toolbar'), _('Editing toolbar'),
                  None, checkable=True),
            'view.addtool':
                a(self, _('Show or hide insert toolbar'), _('Insert toolbar'),
                  None, checkable=True),

            'data.import':
                a(self, _('Import data into Veusz'), _('&Import...'),
                  self.slotDataImport, icon='kde-vzdata-import'),
            'data.edit':
                a(self, _('Edit and enter new datasets'), _('&Editor...'),
                  lambda: self.slotDataEdit(), icon='kde-edit-veuszedit'),
            'data.create':
                a(self, _('Create new datasets using ranges, parametrically or as functions of existing datasets'), _('&Create...'),
                  self.slotDataCreate, icon='kde-dataset-new-veuszedit'),
            'data.create2d':
                a(self, _('Create new 2D datasets from existing datasets, or as a function of x and y'), _('Create &2D...'),
                  self.slotDataCreate2D, icon='kde-dataset2d-new-veuszedit'),
            'data.capture':
                a(self, _('Capture remote data'), _('Ca&pture...'),
                  self.slotDataCapture, icon='veusz-capture-data'),
            'data.filter':
                a(self, _('Filter data'), _('&Filter...'),
                  self.slotDataFilter, icon='kde-filter'),
            'data.histogram':
                a(self, _('Histogram data'), _('&Histogram...'),
                  self.slotDataHistogram, icon='button_bar'),
            'data.reload':
                a(self, _('Reload linked datasets'), _('&Reload'),
                  self.slotDataReload, icon='kde-view-refresh'),

            'help.home':
                a(self, _('Go to the Veusz home page on the internet'),
                  _('Home page'), self.slotHelpHomepage),
            'help.bug':
                a(self, _('Report a bug on the internet'),
                  _('Suggestions and bugs'), self.slotHelpBug),
            'help.update':
                a(self, _('Download latest version'),
                  _('Download latest version'), self.slotHelpUpdate),

            'help.tutorial':
                a(self, _('An interactive Veusz tutorial'),
                  _('Tutorial'), self.slotHelpTutorial),
            'help.about':
                a(self, _('Displays information about the program'), _('About...'),
                  self.slotHelpAbout, icon='veusz')
        }

        # create main toolbar
        tb = self.maintoolbar = qt.QToolBar(_("Main toolbar - Veusz"), self)
        iconsize = setdb['toolbar_size']
        tb.setIconSize(qt.QSize(iconsize, iconsize))
        tb.setObjectName('veuszmaintoolbar')
        self.addToolBar(qt.Qt.TopToolBarArea, tb)

        utils.makeMenuGroupSaved(
            'file.new.menu', self, self.vzactions, (
                'file.new.graph', 'file.new.graph3d',
                'file.new.polar', 'file.new.ternary',
            )
        )

        utils.addToolbarActions(
            tb, self.vzactions,
            ('file.new.menu', 'file.open', 'file.save',
             'file.print', 'file.export'))

        # data toolbar
        tb = self.datatoolbar = qt.QToolBar(_("Data toolbar - Veusz"), self)
        tb.setIconSize(qt.QSize(iconsize, iconsize))
        tb.setObjectName('veuszdatatoolbar')
        self.addToolBar(qt.Qt.TopToolBarArea, tb)
        utils.addToolbarActions(
            tb, self.vzactions,
            ('data.import', 'data.edit',
             'data.create', 'data.capture',
             'data.filter', 'data.reload')
        )

        # menu structure
        filemenu = [
            [
                'file.new', _('New'),
                [
                    'file.new.graph', 'file.new.graph3d', 'file.new.polar',
                    'file.new.ternary'
                ]
            ],
            'file.open',
            [
                'file.filerecent', _('Open &Recent'), []
            ],
            'file.reload',
            '',
            'file.save', 'file.saveas', 'file.trust',
            '',
            'file.print', 'file.export',
            '',
            'file.close', 'file.quit'
        ]
        editmenu = [
            'edit.undo', 'edit.redo',
            '',
            ['edit.select', _('&Select'), []],
            '',
            'edit.prefs', 'edit.stylesheet', 'edit.custom',
            ''
        ]
        viewwindowsmenu = [
            'view.edit', 'view.props', 'view.format',
            'view.console', 'view.datanav',
            '',
            'view.maintool', 'view.viewtool',
            'view.addtool', 'view.edittool'
        ]
        viewmenu = [
            [
                'view.viewwindows', _('&Windows'), viewwindowsmenu
            ],
            ''
        ]
        insertmenu = [
        ]

        # load dataset plugins and create menu
        datapluginsmenu = self.definePlugins(
            plugins.datasetpluginregistry,
            self.vzactions, 'data.ops'
        )

        datamenu = [
            [
                'data.ops', _('&Operations'), datapluginsmenu
            ],
            'data.import', 'data.edit', 'data.create',
            'data.create2d', 'data.capture', 'data.filter', 'data.histogram',
            'data.reload',
        ]
        helpmenu = [
            'help.home', 'help.bug', 'help.update',
            '',
            'help.tutorial',
            '',
            [
                'help.examples', _('&Example documents'), []
            ],
            '',
            'help.about'
        ]

        # load tools plugins and create menu
        toolsmenu = self.definePlugins(
            plugins.toolspluginregistry,
            self.vzactions, 'tools')

        menus = [
            ['file', _('&File'), filemenu],
            ['edit', _('&Edit'), editmenu],
            ['view', _('&View'), viewmenu],
            ['insert', _('&Insert'), insertmenu],
            ['data', _('&Data'), datamenu],
            ['tools', _('&Tools'), toolsmenu],
            ['help', _('&Help'), helpmenu],
        ]

        self.menus = {}
        utils.constructMenus(self.menuBar(), self.menus, menus, self.vzactions)

        # set icon for File->New
        self.menus['file.new'].setIcon(utils.getIcon('kde-document-new'))

        self.populateExamplesMenu()

    def _setPickerFont(self, label):
        f = label.font()
        f.setBold(True)
        f.setPointSizeF(f.pointSizeF() * 1.2)
        label.setFont(f)

    def populateExamplesMenu(self):
        """Add examples to help menu."""

        # not cstr here forces to unicode for Python 2, getting
        # filenames in unicode
        examples = [
            os.path.join(utils.exampleDirectory, f)
            for f in os.listdir(str(utils.exampleDirectory))
            if os.path.splitext(f)[1] == ".vsz"
        ]

        menu = self.menus["help.examples"]
        for ex in sorted(examples):
            name = os.path.splitext(os.path.basename(ex))[0]

            def _openexample(ex=ex):
                MainWindow.CreateWindow(ex)

            a = menu.addAction(name, _openexample)
            a.setStatusTip(_("Open %s example document") % name)

    def defineViewWindowMenu(self):
        """Setup View -> Window menu."""

        def viewHideWindow(window):
            """Toggle window visibility."""
            w = window
            def f():
                w.setVisible(not w.isVisible())
            return f

        # set whether windows are visible and connect up to toggle windows
        self.viewwinfns = []
        for win, act in (
                (self.treeedit, 'view.edit'),
                (self.propdock, 'view.props'),
                (self.formatdock, 'view.format'),
                (self.console, 'view.console'),
                (self.datadock, 'view.datanav'),
                (self.maintoolbar, 'view.maintool'),
                (self.datatoolbar, 'view.datatool'),
                (self.treeedit.edittoolbar, 'view.edittool'),
                (self.treeedit.addtoolbar, 'view.addtool'),
                (self.plot.viewtoolbar, 'view.viewtool')
        ):

            a = self.vzactions[act]
            fn = viewHideWindow(win)
            self.viewwinfns.append( (win, a, fn) )
            a.triggered.connect(fn)

        # needs to update state every time menu is shown
        self.menus['view.viewwindows'].aboutToShow.connect(
            self.slotAboutToShowViewWindow)

    def slotAboutToShowViewWindow(self):
        """Enable/disable View->Window item check boxes."""

        for win, act, fn in self.viewwinfns:
            act.setChecked(not win.isHidden())

    def showDialog(self, dialog):
        """Show dialog given."""
        dialog.dialogFinished.connect(self.deleteDialog)
        self.dialogs.append(dialog)
        dialog.show()
        self.dialogShown.emit(dialog)

    def deleteDialog(self, dialog):
        """Remove dialog from list of dialogs."""
        try:
            idx = self.dialogs.index(dialog)
            del self.dialogs[idx]
        except ValueError:
            pass

    def slotDataImport(self):
        """Display the import data dialog."""
        from ..dialogs import importdialog
        dialog = importdialog.ImportDialog(self, self.document)
        self.showDialog(dialog)
        return dialog

    def slotDataEdit(self, editdataset=None):
        """Edit existing datasets.

        If editdataset is set to a dataset name, edit this dataset
        """
        from ..dialogs import dataeditdialog
        dialog = dataeditdialog.DataEditDialog(self, self.document)
        self.showDialog(dialog)
        if editdataset is not None:
            dialog.selectDataset(editdataset)
        return dialog

    def slotDataCreate(self):
        """Create new datasets."""
        from ..dialogs.datacreate import DataCreateDialog
        dialog = DataCreateDialog(self, self.document)
        self.showDialog(dialog)
        return dialog

    def slotDataCreate2D(self):
        """Create new datasets."""
        from ..dialogs.datacreate2d import DataCreate2DDialog
        dialog = DataCreate2DDialog(self, self.document)
        self.showDialog(dialog)
        return dialog

    def slotDataCapture(self):
        """Capture remote data."""
        from ..dialogs.capturedialog import CaptureDialog
        dialog = CaptureDialog(self.document, self)
        self.showDialog(dialog)
        return dialog

    def slotDataFilter(self):
        """Filter datasets."""
        from ..dialogs.filterdialog import FilterDialog
        dialog = FilterDialog(self, self.document)
        self.showDialog(dialog)
        return dialog

    def slotDataHistogram(self):
        """Histogram data."""
        from ..dialogs.histodata import HistoDataDialog
        dialog = HistoDataDialog(self, self.document)
        self.showDialog(dialog)
        return dialog

    def slotDataReload(self):
        """Reload linked datasets."""
        from ..dialogs.reloaddata import ReloadData
        dialog = ReloadData(self.document, self)
        self.showDialog(dialog)
        return dialog

    def slotHelpHomepage(self):
        """Go to the veusz homepage."""
        qt.QDesktopServices.openUrl(qt.QUrl('https://veusz.github.io/'))

    def slotHelpBug(self):
        """Go to the veusz bug page."""
        qt.QDesktopServices.openUrl(
            qt.QUrl('https://github.com/veusz/veusz/issues') )

    def askTutorial(self):
        """Ask if tutorial wanted."""
        retn = qt.QMessageBox.question(
            self, _("Veusz Tutorial"),
            _("Veusz includes a tutorial to help get you started.\n"
              "Would you like to start the tutorial now?\n"
              "If not, you can access it later through the Help menu."),
            qt.QMessageBox.Yes | qt.QMessageBox.No
        )

        if retn == qt.QMessageBox.Yes:
            self.slotHelpTutorial()

    def slotHelpTutorial(self):
        """Show a Veusz tutorial."""
        if self.document.isBlank():
            # run the tutorial
            from .tutorial import TutorialDock
            tutdock = TutorialDock(self.document, self, self)
            self.addDockWidget(qt.Qt.RightDockWidgetArea, tutdock)
            tutdock.show()
        else:
            # open up a blank window for tutorial
            win = self.CreateWindow()
            win.slotHelpTutorial()

    def slotHelpAbout(self):
        """Show about dialog."""
        from ..dialogs.aboutdialog import AboutDialog
        AboutDialog(self).exec_()

    def askVersionCheck(self, mininterval=2):
        """Check with user whether to do version checks.

        This is only done after the user has been using the program
        for mininterval days

        """

        dayssinceinstall = (
            datetime.date.today() -
            datetime.date(*setting.settingdb['install_date'])).days
        if ( dayssinceinstall<mininterval or
             setting.settingdb['vercheck_asked_user'] or
             setting.settingdb['vercheck_disabled'] or
             utils.disableVersionChecks ):
            return

        retn = qt.QMessageBox.question(
            self, _("Version check"),
            _("Veusz will periodically check for new Veusz versions and\n"
              "let you know if there is a new one available.\n\n"
              "Is this ok? This choice can be changed in Preferences."),
            qt.QMessageBox.Yes | qt.QMessageBox.No,
            qt.QMessageBox.Yes
        )

        setting.settingdb['vercheck_disabled'] = retn==qt.QMessageBox.No
        setting.settingdb['vercheck_asked_user'] = True

    def doVersionCheck(self):
        """Check whether there is a new version.
        """
        self.vzactions['help.update'].setVisible(False)

        # check is done asynchronously
        thread = utils.VersionCheckThread(self)
        thread.newversion.connect(self.slotNewVersion)
        thread.finished.connect(thread.deleteLater)
        thread.start()

    def askFeedbackCheck(self, mininterval=3):
        """Check with user whether to do feedback.

        This is only done after the user has been using the program
        for mininterval days

        """

        dayssinceinstall = (
            datetime.date.today() -
            datetime.date(*setting.settingdb['install_date'])).days
        if ( dayssinceinstall<mininterval or
             setting.settingdb['feedback_asked_user'] or
             setting.settingdb['feedback_disabled'] or
             utils.disableFeedback ):
            return

        retn = qt.QMessageBox.question(
            self, _("Send automatic anonymous feedback"),
            _("Veusz can automatically send anonymous feedback "
              "to the developers, with information about the version "
              "of software dependencies, the computer language and how "
              "often features are used.\n\n"
              "Is this ok? This choice can be changed in Preferences."),
            qt.QMessageBox.Yes | qt.QMessageBox.No,
            qt.QMessageBox.Yes
        )

        setting.settingdb['feedback_disabled'] = retn==qt.QMessageBox.No
        setting.settingdb['feedback_asked_user'] = True

    def doFeedback(self):
        """Give feedback."""
        thread = utils.FeedbackCheckThread(self)
        thread.start()

    def slotNewVersion(self, ver):
        """Called when there is a new version."""
        msg = _('Veusz %s is available for download - see Help menu') % ver
        self.statusBar().showMessage(msg, 5000)
        self.vzactions['help.update'].setText(
            _('Download new Veusz %s') % ver)
        self.vzactions['help.update'].setVisible(True)

    def slotHelpUpdate(self):
        """Open web page to update."""
        qt.QDesktopServices.openUrl(qt.QUrl(
            'https://veusz.github.io/download/'))

    def queryOverwrite(self):
        """Do you want to overwrite the current document.

        Returns qt.QMessageBox.(Yes,No,Cancel)."""

        # include filename in mesage box if we can
        filetext = ''
        if self.filename:
            filetext = " '%s'" % os.path.basename(self.filename)

        return qt.QMessageBox.warning(
            self,
            _("Save file?"),
            _("Document%s was modified. Save first?") % filetext,
            qt.QMessageBox.Save | qt.QMessageBox.Discard |
            qt.QMessageBox.Cancel
        )

    def closeEvent(self, event):
        """Before closing, check whether we need to save first."""

        # if the document has been modified then query user for saving
        if self.document.isModified():
            v = self.queryOverwrite()
            if v == qt.QMessageBox.Cancel:
                event.ignore()
                return
            elif v == qt.QMessageBox.Save:
                self.slotFileSave()

        # store working directory
        setdb['dirname'] = self.dirname

        # store the current geometry in the settings database
        geometry = ( self.x(), self.y(), self.width(), self.height() )
        setdb['geometry_mainwindow'] = geometry

        # store docked windows
        data = self.saveState().data()
        setdb['geometry_mainwindowstate'] = bytes(data)

        # save current setting db
        setdb.writeSettings()

        event.accept()

    def setupWindowGeometry(self):
        """Restoring window geometry if possible."""

        # count number of main windows shown
        nummain = 0
        for w in qt.qApp.topLevelWidgets():
            if isinstance(w, qt.QMainWindow):
                nummain += 1

        # if we can restore the geometry, do so
        if 'geometry_mainwindow' in setdb:
            geometry = setdb['geometry_mainwindow']
            self.resize( qt.QSize(geometry[2], geometry[3]) )
            if nummain <= 1:
                geomrect = qt.QApplication.desktop().availableGeometry()
                newpos = qt.QPoint(geometry[0], geometry[1])
                if geomrect.contains(newpos):
                    self.move(newpos)

        # restore docked window geometry
        if 'geometry_mainwindowstate' in setdb:
            try:
                self.restoreState(setdb['geometry_mainwindowstate'])
            except Exception:
                # type can be wrong if switching between Py2/3 PyQ4/5
                pass

    def slotFileNewGraph(self):
        """New file (graph)."""
        self.CreateWindow()

    def slotFileNewPolar(self):
        """New file (polar)."""
        self.CreateWindow(mode='polar')

    def slotFileNewTernary(self):
        """New file (ternary)."""
        self.CreateWindow(mode='ternary')

    def slotFileNewGraph3D(self):
        """New file (graph3d)."""
        self.CreateWindow(mode='graph3d')

    def slotFileSave(self):
        """Save file."""

        if self.filename == '':
            self.slotFileSaveAs()
        else:
            try:
                with utils.OverrideCursor():
                    ext = os.path.splitext(self.filename)[1]
                    mode = 'hdf5' if ext == '.vszh5' else 'vsz'
                    self.document.save(self.filename, mode)
                    self.updateStatusbar(_("Saved to %s") % self.filename)
            except EnvironmentError as e:
                qt.QMessageBox.critical(
                    self, _("Error - Veusz"),
                    _("Unable to save document as '%s'\n\n%s") %
                    (self.filename, e.strerror))

    def updateTitlebar(self):
        """Put the filename into the title bar."""
        if self.filename == '':
            self.setWindowTitle(_('Untitled - Veusz'))
        else:
            self.setWindowTitle(
                _("%s - Veusz") % os.path.basename(self.filename))

    def plotQueueChanged(self, incr):
        self.plotqueuecount += incr
        text = '•' * self.plotqueuecount
        self.plotqueuelabel.setText(text)

    def fileSaveDialog(self, filters, dialogtitle):
        """A generic file save dialog for exporting / saving.

        filters: list of filters
        """

        fd = qt.QFileDialog(self, dialogtitle)
        fd.setDirectory(self.dirname)
        fd.setFileMode(qt.QFileDialog.AnyFile)
        fd.setAcceptMode(qt.QFileDialog.AcceptSave)
        fd.setNameFilters(filters)

        # selected filetype is saved under a key constructed here
        filetype_re = re.compile(r'.*\(\*\.([a-z0-9]+)\)')
        filtertypes = [filetype_re.match(f).group(1) for f in filters]
        filterkey = '_'.join(['filterdefault'] + filtertypes)
        if filterkey in setting.settingdb:
            filter = setting.settingdb[filterkey]
            if filter in filters:
                fd.selectNameFilter(filter)

        # okay was selected (and is okay to overwrite if it exists)
        if fd.exec_() == qt.QDialog.Accepted:
            # save directory for next time
            self.dirname = fd.directory().absolutePath()
            # update the edit box
            filename = fd.selectedFiles()[0]
            filetype = filetype_re.match(fd.selectedNameFilter()).group(1)
            if os.path.splitext(filename)[1][1:] != filetype:
                filename += '.' + filetype
            setting.settingdb[filterkey] = fd.selectedNameFilter()
            return filename

        return None

    def fileOpenDialog(self, filters, dialogtitle):
        """Display an open dialog and return a filename.

        filters: list of filters in format "Filetype (*.vsz)"
        """

        fd = qt.QFileDialog(self, dialogtitle)
        fd.setDirectory(self.dirname)
        fd.setFileMode( qt.QFileDialog.ExistingFile )
        fd.setAcceptMode( qt.QFileDialog.AcceptOpen )
        fd.setNameFilters(filters)

        # if the user chooses a file
        if fd.exec_() == qt.QDialog.Accepted:
            # save directory for next time
            self.dirname = fd.directory().absolutePath()

            filename = fd.selectedFiles()[0]
            try:
                with open(filename):
                    pass
            except EnvironmentError as e:
                qt.QMessageBox.critical(
                    self, _("Error - Veusz"),
                    _("Unable to open '%s'\n\n%s") %
                    (filename, e.strerror))
                return None
            return filename
        return None

    def slotFileSaveAs(self):
        """Save As file."""

        filters = [_('Veusz document files (*.vsz)')]
        if h5py is not None:
            filters += [_('Veusz HDF5 document files (*.vszh5)')]
        filename = self.fileSaveDialog(filters, _('Save as'))
        if filename:
            self.filename = filename
            self.updateTitlebar()

            self.slotFileSave()

    def openFile(self, filename):
        """Select whether to load the file in the
        current window or in a blank window and calls the appropriate loader"""

        if self.document.isBlank():
            # If the file is new and there are no modifications,
            # reuse the current window
            self.openFileInWindow(filename)
        else:
            # create a new window
            self.CreateWindow(filename)

    def loadDocument(self, filename):
        """Load a Veusz document.
        Return True if loaded ok
        """

        def _callbackunsafe():
            """Callback when loading document to ask whether ok to continue loading
            if unsafe commands are found."""
            return self.checkUnsafe()

        def _callbackimporterror(filename, error):
            """Ask user if they want to give a new filename in case of import
            error.
            """
            with utils.OverrideCursor(qt.Qt.ArrowCursor):
                msgbox = qt.QMessageBox(self)
                msgbox.setWindowTitle(_("Import error"))
                msgbox.setText(
                    _("Could not import data from file '%s':\n\n %s") % (
                        filename, error))
                msgbox.setInformativeText(_("Do you want to look for another file?"))
                msgbox.setStandardButtons(qt.QMessageBox.Yes | qt.QMessageBox.Cancel)
                filename = None
                if msgbox.exec_() == qt.QMessageBox.Yes:
                    filename = qt.QFileDialog.getOpenFileName(self, "Choose data file")
                    filename = filename[0] if filename else None
            return filename

        # save stdout and stderr, then redirect to console
        stdout, stderr = sys.stdout, sys.stderr
        sys.stdout = self.console.con_stdout
        sys.stderr = self.console.con_stderr

        try:
            # get loading mode
            ext = os.path.splitext(filename)[1].lower()
            if ext in ('.vsz', '.py'):
                mode = 'vsz'
            elif ext in ('.h5', '.hdf5', '.he5', '.vszh5'):
                mode = 'hdf5'
            else:
                raise document.LoadError(
                    _("Did not recognise file type '%s'") % ext)

            with utils.OverrideCursor():
                # do the actual loading
                self.document.load(
                    filename,
                    mode=mode,
                    callbackunsafe=_callbackunsafe,
                    callbackimporterror=_callbackimporterror)

        except document.LoadError as e:
            from ..dialogs.errorloading import ErrorLoadingDialog
            if e.backtrace:
                d = ErrorLoadingDialog(self, filename, str(e), e.backtrace)
                d.exec_()
            else:
                qt.QMessageBox.critical(
                    self, _("Error opening %s - Veusz") % filename,
                    str(e))
            return False

        # need to remember to restore stdout, stderr
        sys.stdout, sys.stderr = stdout, stderr

        self.documentsetup = True
        return True

    def openFileInWindow(self, filename):
        """Actually do the work of loading a new document.
        """

        ok = self.loadDocument(filename)
        if not ok:
            return

        # remember file for recent list
        self.addRecentFile(filename)

        # let the main window know
        self.filename = filename
        self.updateTitlebar()
        self.updateStatusbar(_("Opened %s") % filename)

        # use current directory of file if not using cwd mode
        if not setdb['dirname_usecwd']:
            self.dirname = os.path.dirname( os.path.abspath(filename) )

        # notify cmpts which need notification that doc has finished opening
        self.documentOpened.emit()

    def addRecentFile(self, filename):
        """Add a file to the recent files list."""

        recent = setdb['main_recentfiles']
        filename = os.path.abspath(filename)

        if filename in recent:
            del recent[recent.index(filename)]
        recent.insert(0, filename)
        setdb['main_recentfiles'] = recent[:10]
        self.populateRecentFiles()

    def slotFileOpen(self):
        """Open an existing file in a new window."""

        filters = ['*.vsz']
        if h5py is not None:
            filters.append('*.vszh5')

        filename = self.fileOpenDialog(
            [_('Veusz document files (%s)') % ' '.join(filters)],
            _('Open'))
        if filename:
            self.openFile(filename)

    def populateRecentFiles(self):
        """Populate the recently opened files menu with a list of
        recently opened files"""

        def opener(path):
            def _fileOpener():
                self.openFile(path)
            return _fileOpener

        menu = self.menus["file.filerecent"]
        menu.clear()

        if setdb['main_recentfiles']:
            files = [
                f for f in setdb['main_recentfiles'] if os.path.isfile(f)]

            # add each recent file to menu
            newmenuitems = []
            for i, path in enumerate(files):
                newmenuitems.append(
                    ('filerecent%i' % i,_('Open File %s') % path,
                     os.path.basename(path),
                     'file.filerecent', opener(path),
                     '', False, ''))

            menu.setEnabled(True)
            self.recentFileActions = utils.populateMenuToolbars(
                newmenuitems, self.maintoolbar, self.menus)
        else:
            menu.setEnabled(False)

    def slotFileReload(self):
        """Reload document from saved version."""

        retn = qt.QMessageBox.warning(
            self,
            _("Reload file"),
            _("Reload document from file, losing any changes?"),
            qt.QMessageBox.Yes | qt.QMessageBox.Cancel,
            qt.QMessageBox.Cancel
        )
        if retn == qt.QMessageBox.Yes:
            if not os.path.exists(self.filename):
                qt.QMessageBox.critical(
                    self,
                    _("Reload file"),
                    _("File %s no longer exists") % self.filename)
            else:
                self.openFileInWindow(self.filename)

    def slotFileExport(self):
        """Export the graph."""
        from ..dialogs.export import ExportDialog
        dialog = ExportDialog(self, self.document, self.filename)
        self.showDialog(dialog)
        return dialog

    def slotFilePrint(self):
        """Print the document."""
        document.printDialog(self, self.document, filename=self.filename)

    def slotModifiedDoc(self, ismodified):
        """Disable certain actions if document is not modified."""

        # enable/disable file, save menu item
        self.vzactions['file.save'].setEnabled(ismodified)

        # enable/disable reloading from saved document
        self.vzactions['file.reload'].setEnabled(
            bool(self.filename) and ismodified)

    def slotFileClose(self):
        """File close window chosen."""
        self.close()

    def slotFileQuit(self):
        """File quit chosen."""
        qt.qApp.closeAllWindows()

    def slotUpdatePage(self, number):
        """Update page number when the plot window says so."""

        nump = self.document.getNumberPages()
        if nump == 0:
            self.pagelabel.setText(_("No pages"))
        else:
            self.pagelabel.setText(_("Page %i/%i") % (number+1, nump))

    def slotUpdateAxisValues(self, values):
        """Update the position where the mouse is relative to the axes."""

        if values:
            # construct comma separated text representing axis values
            valitems = [
                '%s=%#.4g' % (name, values[name])
                for name in sorted(values) ]
            self.axisvalueslabel.setText(', '.join(valitems))
        else:
            self.axisvalueslabel.setText(_('No position'))

    def slotPickerEnabled(self, enabled):
        if enabled:
            self.pickerlabel.setText(_('No point selected'))
            self.pickerlabel.show()
        else:
            self.pickerlabel.hide()

    def slotUpdatePickerLabel(self, info):
        """Display the picked point"""
        xv, yv = info.coords
        xn, yn = info.labels
        xt, yt = info.displaytype
        ix = str(info.index)
        if ix:
            ix = '[' + ix + ']'

        # format values for display
        def fmt(val, dtype):
            if dtype == 'date':
                return utils.dateFloatToString(val)
            elif dtype == 'numeric':
                fmt = '%.'+str(setting.settingdb['picker_sig_figs'])+'g'
                return fmt % val
            elif dtype == 'text':
                return val
            else:
                raise RuntimeError

        xtext = fmt(xv, xt)
        ytext = fmt(yv, yt)

        t = '%s: %s%s = %s, %s%s = %s' % (
            info.widget.name, xn, ix, xtext, yn, ix, ytext)
        self.pickerlabel.setText(t)
        if setdb['picker_to_console']:
            self.console.appendOutput(t + "\n", 'error')
        if setdb['picker_to_clipboard']:
            clipboard = qt.QApplication.clipboard()
            if clipboard.mimeData().hasText():
                clipboard.setText(clipboard.text()+"\n"+t)
            else:
                qt.QApplication.clipboard().setText(t)

    def checkUnsafe(self):
        """Ask user if code should be unsafe."""

        # we shouldn't allow these places to be added to allowed safe directories
        badlocs = (
            qt.QStandardPaths.standardLocations(qt.QStandardPaths.DownloadLocation) +
            qt.QStandardPaths.standardLocations(qt.QStandardPaths.TempLocation)
        )
        fname = self.document.filename
        absfile = os.path.abspath(fname)
        filedir = '' if not fname else os.path.dirname(absfile)
        isbadloc = False
        if not filedir:
            # shouldn't get here, but don't allow empty dir to be added
            isbadloc = True
        for path in badlocs:
            if absfile.startswith(os.path.abspath(path) + os.sep):
                isbadloc = True
        # don't allow plain home directory
        if filedir == qt.QStandardPaths.standardLocations(
                qt.QStandardPaths.HomeLocation)[0]:
            isbadloc = True

        msgbox = qt.QMessageBox(
            qt.QMessageBox.Warning,
            _("Potentially unsafe code in document"),
            _(
                "<p><b>The document '%s' contains potentially unsafe code</b></p>"
                "<p>Directory: '%s'</p>"
                "<p>This file could damage your computer or data "
                "as it can contain arbitrary code. "
                "Please check "
                "that the file was made by you or a trusted source.</p>"
            ) % (
                os.path.basename(absfile) if fname else "",
                filedir if fname else "",
            ),
            qt.QMessageBox.NoButton,
            self,
        )
        allow = msgbox.addButton(_("Allow"), qt.QMessageBox.AcceptRole)
        addloc = msgbox.addButton(_("Add to trusted locations"), qt.QMessageBox.AcceptRole)
        addloc.setEnabled(not isbadloc)
        stop = msgbox.addButton(_("Skip"), qt.QMessageBox.RejectRole)

        msgbox.setDefaultButton(stop)

        # we enter here with a busy cursor often, so set back to arrow temporarily
        with utils.OverrideCursor(qt.Qt.ArrowCursor):
            msgbox.exec_()

        clicked = msgbox.clickedButton()
        if clicked is addloc:
            with utils.OverrideCursor(qt.Qt.ArrowCursor):
                button = qt.QMessageBox.warning(
                    self, _("Are you sure?"),
                    _("Are you really sure that you want to add directory '%s' to the "
                      "list of trusted locations. Any file loaded from this directory "
                      "will be trusted.") % filedir,
                    qt.QMessageBox.Yes | qt.QMessageBox.No,
                    qt.QMessageBox.No,
                )
            if button == qt.QMessageBox.Yes:
                setting.settingdb['secure_dirs'].append(filedir)
                return True
            else:
                return False

        return clicked is allow

    def slotUpdateSecurity(self, secure):
        """Show or hide security label and trust menu based on security"""
        self.securitylabel.setVisible(not secure)
        self.vzactions['file.trust'].setVisible(not secure)

    def slotAllowedImportsDoc(self):
        """Are allowed imports?"""
        if self.checkUnsafe():
            self.document.evaluate.setSecurity(True)

    def slotFileTrust(self):
        """User requests that document should be trusted."""
        button = qt.QMessageBox.warning(
            self, _("Are you sure?"),
            _("Are you sure that you want to trust the document contents, "
              "including any potentially dangerous code? Only trust "
              "documents with a trusted source."),
            qt.QMessageBox.Yes | qt.QMessageBox.No,
            qt.QMessageBox.No,
        )
        if button == qt.QMessageBox.Yes:
            self.document.evaluate.setSecurity(True)
            # force redraw of document
            self.plot.actionForceUpdate()



#    Copyright (C) 2004 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
##############################################################################

"""SimpleRead: a class for the reading of data formatted in a simple way

To read the data it takes a descriptor which takes the form of

varname<error specifiers><[repeater]>
where <> marks optional arguments, e.g.

x+- y+,- z+-[1:5]
x(text) y(date) z(number),+-

+- means symmetric error bars
+,- means asymmetric error bars
, is a separator
(text) or (date) specifies datatype

z+-[1:5] means read z_1+- z_2+- ... z_5+-

Indicices can be unspecified, [:], [1:], [:5]
The first 3 mean the same thing, the last means read from 1 to 5

Commas are now optional in 1.6, so descriptors can look like
x +- y + -
"""

import re
import ast
import io
import numpy as N

from .. import utils
from .. import datasets
from .. import qtall as qt
from .import base

# a regular expression for splitting descriptor into tokens
descrtokens_split_re = re.compile(r'''
(
 `[^`]*`       |  # quoted name
 [ ,]          |  # comma or space
 \([a-z]+?\)   |  # data type
 \+- | \+ | -  |  # error bars
 \[.*?\]          # indices
)
''', re.VERBOSE)

range_re = re.compile(r'''^
 \[
 (-?[0-9]+)?
 :
 (-?[0-9]+)?
 \]
$''', re.VERBOSE)

def interpretDescriptor(descr):
    """Get a descriptor and create a set of descriptor objects."""

    parts = []

    split = descrtokens_split_re.split(descr.strip())
    tokens = [x for x in split if x != '']
    # make sure that last dataset is added
    tokens += ['DUMMY']

    name = datatype = idxrange = None
    columns = []
    for tokenindex, token in enumerate(tokens):
        # skip spaces
        if token == ' ':
            if tokenindex > 0 and tokens[tokenindex-1] == ',':
                columns.append(',')
            continue

        # ignore column
        if token == ',':
            if tokenindex == 0 or (
                    tokens[tokenindex-1] == ',' or
                    tokens[tokenindex-1] == ' ' ):
                columns.append(',')
            continue

        # does token match datatype name?
        if ( token[0] == '(' and token[-1] == ')' and
             token[1:-1] in datatype_name_convert ):
            datatype = datatype_name_convert[token[1:-1]]
            continue

        # match error bars
        if token in ('+', '-', '+-'):
            columns.append(token)
            continue

        # does token match a range?
        m = range_re.match(token)
        if m:
            if m.group(1):
                startindex = int(m.group(1))
            else:
                startindex = 1
            if m.group(2):
                stopindex = int(m.group(2))
            else:
                stopindex = 999999999
            idxrange = (startindex, stopindex)
            continue

        # quoted dataset name, so remove quotes
        if token[0] == '`' and token[-1] == '`':
            token = token[1:-1]

        # add previous entry
        if name is not None:
            parts.append( DescriptorPart(name, datatype, columns, idxrange) )
            name = datatype = idxrange = None
            columns = []

        columns.append('D')
        name = token

    return parts

class DescriptorError(ValueError):
    """Used to indicate an error with the descriptor."""
    pass

# this is a regular expression to match properly quoted strings
# hopefully a matching expression can be passed to eval
string_re = re.compile( r'''
^
u?"" |            # match empty double-quoted string
u?".*?[^\\]" |    # match double-quoted string, ignoring escaped quotes
u?'' |            # match empty single-quoted string
u?'.*?[^\\]'      # match single-quoted string, ignoring escaped quotes
$
''', re.VERBOSE )

# a line starting with text
text_start_re = re.compile( r'^[A-Za-z]' )

# convert data type strings in descriptor to internal datatype
datatype_name_convert = {
    'float': 'float',
    'numeric': 'float',
    'number': 'float',
    'text': 'string',
    'string': 'string',
    'date': 'date',
    'time': 'date'
}

def guessDataType(val):
    """Try to work out data type from sample value (val)

    Return values are one of
    float, string, or date
    """

    # if the dataset type is specified
    # check for identifiers in dataset name
    # guess the type:
    # obvious float
    try:
        float(val)
        return 'float'
    except ValueError:
        pass

    # do all libcs check for these?
    if val.lower() in ('inf', '+inf', '-inf', 'nan'):
        return 'float'

    # obvious string
    if string_re.match(val):
        return 'string'

    # date
    if utils.isDateTime(val):
        return 'date'

    # assume string otherwise
    return 'string'

class DescriptorPart:
    """Represents part of a descriptor."""

    def __init__(self, name, datatype, columns, idxrange):
        """Construct DescriptorPart
        name is dataset name
        datatype is None or one of the possible options
        columns is a list of the columns '+', '-', '+-', ',' or 'D'
         for errors, ignoring a column or a data column
        idxrange is None or a tuple (minidx, maxidx)
        """
        self.name = name
        self.datatype = datatype
        self.columns = tuple(columns)

        self.errorcount = 0

        self.single = idxrange is None
        if self.single:
            self.startindex = self.stopindex = 1
        else:
            self.startindex, self.stopindex = idxrange

    def readFromStream(self, stream, thedatasets, block=None):
        """Read data from stream, and write to thedatasets."""

        # loop over column range
        for index in range(self.startindex, self.stopindex+1):
            # name for variable
            if self.single:
                name = self.name
            else:
                name = '%s_%i' % (self.name, index)

            # if we're reading multiple blocks
            if block is not None:
                name += '_%i' % block

            # loop over columns until we run out, or we don't need any
            for col in self.columns:
                # get next column and return if we run out of data
                val = stream.nextColumn()
                if val is None:
                    return
                # append a suffix to specify whether error or value
                # \0 is used as the user cannot enter it
                fullname = '%s\0%s' % (name, col)

                # get dataset (or get new one)
                try:
                    dataset = thedatasets[fullname]
                except KeyError:
                    dataset = thedatasets[fullname] = []

                if not self.datatype:
                    # try to guess type of data
                    self.datatype = guessDataType(val)

                # convert according to datatype
                if self.datatype == 'float':
                    try:
                        # do conversion
                        dat = float(val)
                    except ValueError:
                        dat = N.nan
                        self.errorcount += 1

                elif self.datatype == 'string':
                    if string_re.match(val):
                        conv = val
                        if conv[0:1] != 'u':
                            # a hack for python2
                            conv = 'u' + val
                        try:
                            dat = ast.literal_eval(conv)
                        except:
                            dat = val
                    else:
                        dat = val

                elif self.datatype == 'date':
                    dat = utils.dateStringToDate(val)

                # add data into dataset
                dataset.append(dat)

    def setOutput(self, thedatasets, outmap, block=None,
                  linkedfile=None,
                  prefix="", suffix="", tail=None):
        """Set the read-in data in the document."""

        # we didn't read any data
        if self.datatype is None:
            return

        for index in range(self.startindex, self.stopindex+1):
            # name for variable
            if self.single:
                name = '%s' % (self.name,)
            else:
                name = '%s_%i' % (self.name, index)
            if block is not None:
                name += '_%i' % block

            # does the dataset exist?
            if name+'\0D' in thedatasets:
                vals = thedatasets[name+'\0D']
                pos = neg = sym = None

                # retrieve the data for this dataset
                if name+'\0+' in thedatasets: pos = thedatasets[name+'\0+']
                if name+'\0-' in thedatasets: neg = thedatasets[name+'\0-']
                if name+'\0+-' in thedatasets: sym = thedatasets[name+'\0+-']

                # make sure components are the same length
                minlength = 99999999999999
                for ds in vals, pos, neg, sym:
                    if ds is not None and len(ds) < minlength:
                        minlength = len(ds)
                for ds in vals, pos, neg, sym:
                    if ds is not None and len(ds) != minlength:
                        del ds[minlength:]

                # only remember last N values
                if tail is not None:
                    vals = vals[-tail:]
                    if sym is not None: sym = sym[-tail:]
                    if pos is not None: pos = pos[-tail:]
                    if neg is not None: neg = neg[-tail:]

                # create the dataset
                if self.datatype == 'float':
                    ds = datasets.Dataset( data = vals, serr = sym,
                                           nerr = neg, perr = pos,
                                           linked = linkedfile )
                elif self.datatype == 'date':
                    ds = datasets.DatasetDateTime( data=vals,
                                                   linked=linkedfile )
                elif self.datatype == 'string':
                    ds = datasets.DatasetText( data=vals,
                                               linked = linkedfile )
                else:
                    raise RuntimeError("Invalid data type")

                finalname = prefix + name + suffix
                outmap[finalname] = ds
            else:
                break

class Stream:
    """This object reads through an input data source (override
    readLine) and interprets data from the source."""

    # this is a regular expression for finding data items in data stream
    # I'll try to explain this bit-by-bit (these are ORd, and matched in order)
    find_re = re.compile( r'''
    `.+?`[^ \t\n\r#!%;]* | # match dataset name quoted in back-ticks
                           # we also need to match following characters to catch
                           # corner cases in the descriptor
    u?"" |          # match empty double-quoted string
    u?".*?[^\\]" |  # match double-quoted string, ignoring escaped quotes
    u?'' |          # match empty single-quoted string
    u?'.*?[^\\]' |  # match single-quoted string, ignoring escaped quotes
    [#!%;](?=descriptor) | # match separately comment char before descriptor
    [#!%;].* |      # match comment to end of line
    [^ \t\n\r#!%;]+ # match normal space/tab separated items
    ''', re.VERBOSE )

    def __init__(self):
        """Initialise stream object."""
        self.remainingline = []

    def nextColumn(self):
        """Return value of next column of line."""
        try:
            return self.remainingline.pop(0)
        except IndexError:
            return None

    def allColumns(self):
        """Get all columns of current line (none are discarded)."""
        return self.remainingline

    def flushLine(self):
        """Forget the rest of the line."""
        self.remainingline = []

    def readLine(self):
        """Read the next line of the data source.
        StopIteration is raised if there is no more data."""
        pass

    def newLine(self):
        """Read in, and split the next line."""

        while True:
            # get next line from data source
            try:
                line = self.readLine()
            except StopIteration:
                # end of file
                return False

            # break up and append to buffer (removing comments)
            cmpts = self.find_re.findall(line)
            self.remainingline += [ x for x in cmpts if x[0] not in '#!%;']

            if self.remainingline and self.remainingline[-1] == '\\':
                # this is a continuation: drop this item and read next line
                self.remainingline.pop()
            else:
                return True

class FileStream(Stream):
    """A stream based on a python-style file (or iterable)."""

    def __init__(self, file):
        """File can be any iterator-like object."""
        Stream.__init__(self)
        self.file = file

    def readLine(self):
        """Read the next line of the data source.
        StopIteration is raised if there is no more data."""
        return next(self.file)

class StringStream(FileStream):
    '''For reading data from a string.'''

    def __init__(self, text):
        """A stream which reads in from a text string."""

        FileStream.__init__( self, io.StringIO(text) )

class CSVStream(Stream):
    """Read text from csv file."""

    def __init__(self, filename, delim, textdelim, locale, encoding):
        Stream.__init__(self)

        self.csvfile = utils.get_unicode_csv_reader(
            filename,
            delimiter=delim,
            quotechar=textdelim,
            encoding=encoding )
        self.localename = locale
        self.locale = qt.QLocale(locale)

    def newLine(self):
        """Get next line from CSV file."""
        try:
            line = next(self.csvfile)
        except StopIteration:
            return False

        # delete empty cells on left, to make compatible with normal
        # text stream
        i = 0
        while i < len(line) and not line[i]:
            i += 1
        line = line[i:]

        if self.localename == 'en_US':
            # no conversion
            self.remainingline += line
        else:
            for t in line:
                v, ok = self.locale.toDouble(t)
                if ok:
                    # add on converted text - yuck - double conversion
                    self.remainingline.append('%e' % v)
                else:
                    self.remainingline.append(t)
        return True

class SimpleRead:
    '''Class to read in datasets from a stream.

    The descriptor specifies the format of data to read from the stream
    Read the docstring for this module for information

    tail attribute if set says to only use last tail data points when setting
    '''

    def __init__(self, descriptor):
        # convert descriptor to part objects
        descriptor = descriptor.strip()
        self._parseDescriptor(descriptor)

        # construct data names automatically
        self.autodescr = (descriptor == '')

        # get read for reading data
        self.clearState()

    def clearState(self):
        """Start reading from scratch."""
        self.datasets = {}
        self.blocks = None
        self.tail = None

    def _parseDescriptor(self, descriptor):
        """Take a descriptor, and parse it into its individual parts."""
        self.parts = interpretDescriptor(descriptor)

    def readData(self, stream, useblocks=False, ignoretext=False):
        """Read in the data from the stream.

        If useblocks is True, data are read as separate blocks.
        Dataset names are appending with an underscore and a block
        number if set.
        """

        self.ignoretext = ignoretext
        if useblocks:
            self._readDataBlocked(stream, ignoretext)
        else:
            self._readDataUnblocked(stream, ignoretext)

    def _readDataUnblocked(self, stream, ignoretext):
        """Read in that data from the stream."""

        allparts = list(self.parts)

        # loop over lines
        while stream.newLine():
            if stream.remainingline[:1] == ['descriptor']:
                # a change descriptor statement
                descriptor =  ' '.join(stream.remainingline[1:])
                self._parseDescriptor(descriptor)
                allparts += self.parts
                self.autodescr = False
            elif ( self.ignoretext and len(stream.remainingline) > 0 and
                   text_start_re.match(stream.remainingline[0]) and
                   len(self.parts) > 0 and
                   self.parts[0].datatype != 'string' and
                   stream.remainingline[0] not in ('inf', 'nan') ):
                # ignore the line if it is text and ignore text is on
                # and first column is not text
                pass
            else:
                # normal text
                for p in self.parts:
                    p.readFromStream(stream, self.datasets)

                # automatically create parts if data are remaining
                if self.autodescr:
                    while len(stream.remainingline) > 0:
                        p = DescriptorPart(
                            str(len(self.parts)+1), None, 'D', None )
                        p.readFromStream(stream, self.datasets)
                        self.parts.append(p)
                        allparts.append(p)

            stream.flushLine()

        self.parts = allparts
        self.blocks = None

    def _readDataBlocked(self, stream, ignoretext):
        """Read in the data, using blocks."""

        allparts = list(self.parts)

        blocks = {}
        block = 1
        while stream.newLine():
            line = stream.remainingline

            # if this is a blank line, separating data then advance to a new
            # block
            if len(line) == 0 or line[0].lower() == 'no':
                # blank lines separate blocks
                if block in blocks:
                    block += 1
            else:
                # read in data
                for p in self.parts:
                    p.readFromStream(stream, self.datasets, block=block)

                # automatically create parts if data are remaining
                if self.autodescr:
                    while len(stream.remainingline) > 0:
                        p = DescriptorPart(
                            str(len(self.parts)+1), None, 'D', None )
                        p.readFromStream(stream, self.datasets, block=block)
                        self.parts.append(p)
                        allparts.append(p)

                blocks[block] = True

            # lose remaining data
            stream.flushLine()

        self.parts = allparts
        self.blocks = list(blocks.keys())

    def getInvalidConversions(self):
        """Return the number of invalid conversions after reading data.

        Returns a dict of dataset, number values."""

        out = {}
        for p in self.parts:
            out[p.name] = p.errorcount
        return out

    def getDatasetCounts(self):
        """Get a dict of the datasets read (main data part) and number
        of entries read."""
        out = {}
        for name in self.datasets:
            if name[-2:] == '\0D':
                out[name[:-2]] = len(self.datasets[name])
        return out

    def setOutput(self, out, linkedfile=None,
                  prefix='', suffix=''):
        """Set the data in the out dict.
        """

        # iterate over blocks used
        if self.blocks is None:
            blocks = [None]
        else:
            blocks = self.blocks

        # if automatically making parts, use a prefix/suffix if not set
        if self.autodescr and prefix == '' and suffix == '':
            prefix = 'col'

        for block in blocks:
            for part in self.parts:
                part.setOutput(
                    self.datasets, out,
                    block=block,
                    linkedfile=linkedfile,
                    prefix=prefix, suffix=suffix,
                    tail=self.tail)

#####################################################################
# 2D data reading

class Read2DError(base.ImportingError):
    pass

class SimpleRead2D:
    def __init__(self, name, params):
        """Read dataset with name given.
        params is a ImportParams2D object
        """
        self.name = name
        self.params = params.copy()

        # not present in ImportParams2D
        self.xedge = self.yedge = self.xcent = self.ycent = None

    ####################################################################
    # Follow functions are for setting parameters during reading of data

    def _paramXRange(self, cols):
        try:
            self.params.xrange = ( float(cols[1]), float(cols[2]) )
        except ValueError:
            raise Read2DError("xrange is not two numerical values")

    def _paramYRange(self, cols):
        try:
            self.params.yrange = ( float(cols[1]), float(cols[2]) )
        except ValueError:
            raise Read2DError("yrange is not two numerical values")

    def _getNumList(self, attr, cols):
        """Generic conversion of a list of numbers."""
        try:
            g = [float(v) for v in cols[1:]]
        except ValueError:
            raise Read2DError("%s is not a list of numerical values" % attr)
        if utils.checkOrder(g) == 0:
            raise Read2DError("%s are not ascending or descending" % attr)
        setattr(self, attr, g)

    def _paramInvertRows(self, cols):
        self.params.invertrows = True

    def _paramInvertCols(self, cols):
        self.params.invertcols = True

    def _paramTranspose(self, cols):
        self.params.transpose = True

    def _paramGridAtEdge(self, cols):
        self.params.gridatedge = True

    ####################################################################

    def readData(self, stream):
        """Read data from stream given

        stream consists of:
        optional:
         xrange A B   - set the range of x from A to B
         yrange A B   - set the range of y from A to B
         xedge A B... - list of x values (instead of xrange)
         yedge A B... - list of y values (instead of yrange)
         xcent A B... - list of x centres (instead of xrange)
         ycent A B... - list of y centres (instead of yrange)
         invertrows   - invert order of the rows
         invertcols   - invert order of the columns
         transpose    - swap rows and columns
         gridatedge   - positions of pixels are given at top and left
        then:
         matrix of columns and rows, separated by line endings
         the rows are in reverse-y order (highest y first)
         blank line stops reading for further datasets
        """

        settings = {
            'xrange': self._paramXRange,
            'yrange': self._paramYRange,
            'xedge': lambda cols: self._getNumList('xedge', cols),
            'yedge': lambda cols: self._getNumList('yedge', cols),
            'xcent': lambda cols: self._getNumList('xcent', cols),
            'ycent': lambda cols: self._getNumList('ycent', cols),
            'invertrows': self._paramInvertRows,
            'invertcols': self._paramInvertCols,
            'transpose': self._paramTranspose,
            'gridatedge': self._paramGridAtEdge,
        }

        rows = []
        # loop over lines
        while stream.newLine():
            cols = stream.allColumns()

            if len(cols) == 0:
                if len(rows) != 0:
                    # end of data
                    break
                continue

            # check to see whether parameter is set
            c = cols[0].lower()
            if c in settings:
                settings[c](cols)
                stream.flushLine()
                continue

            # read columns
            line = []
            while True:
                v = stream.nextColumn()
                if v is None:
                    break
                try:
                    line.append( float(v) )
                except ValueError:
                    raise Read2DError("Could not interpret number '%s'" % v)

            rows.insert(0, line)

        if self.params.gridatedge:

            if any( [getattr(self, x) is not None
                     for x in ("xedge", "yedge", "xcent", "ycent")] ):
                raise Read2DError(
                    "x|y grid|cent are incompatible with gridatedge")

            if len(rows) > 0:
                self.xcent = N.array(rows[-1])
            else:
                self.xcent = N.array([])
            self.ycent = N.array([r[0] for r in rows[:-1]])

            # chop out grid
            rows = [ r[1:] for r in rows[:-1] ]

        # dodgy formatting probably...
        if len(rows) == 0:
            raise Read2DError("No data could be imported for dataset")

        try:
            self.data = N.array(rows, dtype=N.float64)
        except ValueError:
            raise Read2DError("Could not convert data to 2D matrix")

        # obvious check
        if len(self.data.shape) != 2:
            raise Read2DError("Dataset was not 2D")

        if self.params.invertcols:
            self.data = self.data[:,::-1]
        if self.params.invertrows:
            self.data = self.data[::-1,:]

        # transpose matrix if requested
        if self.params.transpose:
            self.data = N.transpose(self.data).copy()
            self.xedge, self.yedge = self.xedge, self.yedge

        # check orders of coords - flip if wrong
        for attr in 'xedge', 'xcent', 'yedge', 'ycent':
            v = getattr(self, attr)
            if v is not None:
                order = utils.checkOrder(v)
                if order == 0:
                    raise Read2DError(
                        '%s must be ascending or descending' % attr)
                elif order == -1:
                    # flip direction of coord and data
                    setattr(self, attr, v[::-1])
                    if attr[0] == 'x':
                        self.data = self.data[:,::-1]
                    else:
                        self.data = self.data[::-1,:]

        # more sanity checks
        if ( (self.xedge is not None and
              len(self.xedge) != self.data.shape[1]+1) or
             (self.yedge is not None and
              len(self.yedge) != self.data.shape[0]+1) ):
            raise Read2DError("xedge and yedge lengths must be data shape+1")

        if ( (self.xcent is not None and
              len(self.xcent) != self.data.shape[1]) or
             (self.ycent is not None and
              len(self.ycent) != self.data.shape[0]) ):
            raise Read2DError("xcent and ycent lengths must be data shape")

    def setOutput(self, out, linkedfile=None):
        """Set the data in the output dict out
        """

        ds = datasets.Dataset2D(
            self.data,
            xrange=self.params.xrange,
            yrange=self.params.yrange,
            xedge=self.xedge, yedge=self.yedge,
            xcent=self.xcent, ycent=self.ycent
        )

        ds.linked = linkedfile

        fullname = self.params.prefix + self.name + self.params.suffix
        out[fullname] = ds

#####################################################################
# n-dimensional data reading

class ReadNDError(base.ImportingError):
    pass

class SimpleReadND:
    def __init__(self, name, params):
        """Read dataset with name given.
        params is a ImportParamsND object
        """
        self.name = name
        self.params = params.copy()

    ####################################################################

    def _paramTranspose(self, cols):
        self.params.transpose = True

    def _paramShape(self, cols):
        sizes = [int(x) for x in cols[1:]]
        self.params.shape = tuple(sizes)

    ####################################################################

    def readData(self, stream):
        """Read data from stream given

        stream consists of:
        optional:
         transpose    - swap rows and columns
         shape size1 size2 ... - give dimensions of dataset
        then:
         Matrix of columns and rows, separated by line endings.
         A single line is a single dimension
         Higher orders are given by using increasing numbers of
         separating newlines
        """

        settings = {
            'transpose': self._paramTranspose,
            'shape': self._paramShape,
        }

        vals = []

        # keep track of where we are in terms of index
        dimstack = []
        dimidx = 0

        # loop over lines
        while stream.newLine():
            cols = stream.allColumns()

            # check to see whether parameter is set
            if len(cols) > 0:
                c = cols[0].lower()
                if c in settings:
                    settings[c](cols)
                    stream.flushLine()
                    continue

            # read columns
            line = []
            while True:
                v = stream.nextColumn()
                if v is None:
                    break
                try:
                    line.append(float(v))
                except ValueError:
                    raise ReadNDError("Could not interpret number '%s'" % v)

            if len(line) > 0:
                # previous blank lines
                if dimidx != len(dimstack):
                    # insert new dimensions, if required
                    while dimidx < 0:
                        dimstack.insert(0, 0)
                        dimidx += 1
                        vals = [vals]

                    # move to next value at current level
                    dimstack[dimidx] += 1
                    for i in range(dimidx+1, len(dimstack)):
                        dimstack[i] = 0
                    dimidx = len(dimstack)

                # lookup correct place in hierarchy and append
                v = vals
                for s in dimstack:
                    while s >= len(v):
                        v.append([])
                    v = v[s]
                v.append(N.array(line, dtype=N.float64))
            else:
                if len(vals) > 0:
                    dimidx -= 1

        if self.params.shape is not None:
            # flatten so we can reshape properly later (this is to
            # allow free form input with the shape option)
            fdata = []
            def flatten(d):
                for x in d:
                    if isinstance(x, list):
                        flatten(x)
                    else:
                        fdata.append(x)
            flatten(vals)
            vals = N.hstack((fdata))

        try:
            self.data = N.array(vals, dtype=N.float64)
        except ValueError:
            raise ReadNDError("Could not convert data to n-D matrix")

        # obvious check
        if self.data.ndim < 1:
            raise ReadNDError("Needs at least a 1D dataset")

        # for 1d data, an extra dimension is sometimes added
        if self.data.ndim == 2 and self.data.shape[0] == 1:
            self.data = self.data.reshape(self.data.shape[-1])

        # reshape if requested
        if self.params.shape is not None:
            self.data = self.data.reshape(self.params.shape)

        # transpose matrix if requested
        if self.params.transpose:
            self.data = N.transpose(self.data).copy()

    def setOutput(self, out, linkedfile=None):
        """Set the data in the output dict out
        """

        ds = datasets.DatasetND(self.data)
        ds.linked = linkedfile
        fullname = self.params.prefix + self.name + self.params.suffix
        out[fullname] = ds



#    Copyright (C) 2005 Jeremy S. Sanders
#    Email: Jeremy Sanders <jeremy@jeremysanders.net>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
###############################################################################

"""Module for holding collections of settings."""

from .reference import Reference, ReferenceMultiple

class Settings:
    """A class for holding collections of settings."""

    # differentiate widgets, settings and setting
    nodetype = 'settings'

    # various items in class hierarchy
    iswidget = False
    issetting = False
    issettings = True

    def __init__(self, name, descr = '', usertext='', pixmap='',
                 setnsmode='formatting'):
        """A new Settings with a name.

        name: name in hierarchy
        descr: description (for user)
        usertext: name for user of class
        pixmap: pixmap to show in tab (if appropriate)
        setnsmode: type of Settings class, one of
              ('formatting', 'groupedsetting', 'widgetsettings', 'stylesheet',
               'hide')
        """

        self.__dict__['setdict'] = {}
        self.name = name
        self.descr = descr
        self.usertext = usertext
        self.pixmap = pixmap
        self.setnsmode = setnsmode
        self.setnames = []  # a list of names
        self.parent = None

    def copy(self):
        """Make a copy of the settings and its subsettings."""

        s = Settings(
            self.name, descr=self.descr, usertext=self.usertext,
            pixmap=self.pixmap, setnsmode=self.setnsmode )
        for name in self.setnames:
            s.add( self.setdict[name].copy() )
        return s

    def getList(self):
        """Get a list of setting or settings types."""
        return [self.setdict[n] for n in self.setnames]

    def getSettingList(self):
        """Get a list of setting types."""
        return [
            self.setdict[n] for n in self.setnames
            if not isinstance(self.setdict[n], Settings)
        ]

    def getSettingsList(self):
        """Get a list of settings types."""
        return [
            self.setdict[n] for n in self.setnames
            if isinstance(self.setdict[n], Settings)
        ]

    def getNames(self):
        """Return list of names."""
        return self.setnames

    def getSettingNames(self):
        """Get list of setting names."""
        return [
            n for n in self.setnames
            if not isinstance(self.setdict[n], Settings)
        ]

    def getSettingsNames(self):
        """Get list of settings names."""
        return [
            n for n in self.setnames
            if isinstance(self.setdict[n], Settings)
        ]

    def isSetting(self, name):
        """Is the name a supported setting?"""
        return name in self.setdict

    def add(self, setting, posn=-1, readonly=False, pixmap=None):
        """Add a new setting with the name, or a set of subsettings."""
        name = setting.name
        if name in self.setdict:
            raise RuntimeError("Name already in settings dictionary")
        self.setdict[name] = setting
        if posn < 0:
            self.setnames.append(name)
        else:
            self.setnames.insert(posn, name)
        setting.parent = self

        if pixmap:
            setting.pixmap = pixmap

        if readonly:
            setting.readonly = True

    def remove(self, name):
        """Remove name from the list of settings."""

        del self.setnames[ self.setnames.index( name ) ]
        del self.setdict[ name ]

    def __setattr__(self, name, val):
        """Allow us to do

        foo.setname = 42
        """

        d = self.__dict__['setdict']
        if name in d:
            d[name].val = val
        else:
            self.__dict__[name] = val

    def __getattr__(self, name):
        """Allow us to do

        print foo.setname
        """

        try:
            s = self.__dict__['setdict'][name]
            if isinstance(s, Settings):
                return s
            return s.val
        except KeyError:
            pass

        try:
            return self.__dict__[name]
        except KeyError:
            raise AttributeError("'%s' is not a setting" % name)

    def __getitem__(self, name):
        """Also allows us to do

        print foo['setname']
        """

        d = self.__dict__['setdict']
        try:
            s = d[name]
            if isinstance(s, Settings):
                return s
            else:
                return s.val
        except KeyError:
            raise KeyError("'%s' is not a setting" % name)

    def __contains__(self, name):
        """Whether settings contains name."""
        return name in self.__dict__['setdict']

    def get(self, name = None):
        """Get the setting variable."""

        if name is None:
            return self
        else:
            return self.setdict[name]

    def saveText(self, saveall, rootname = None):
        """Return the text which would reload the settings.

        if saveall is true, save those which haven't been modified.
        rootname is the part to stick on the front of the settings
        """

        # we want to build the root up if we're not the first item
        # (first item is implicit)
        if rootname is None:
            rootname = ''
        else:
            rootname += self.name + '/'

        text = ''.join( [
            self.setdict[name].saveText(saveall, rootname)
            for name in self.setnames
        ] )
        return text

    def linkToStylesheet(self, _root=None):
        """Link the settings within this Settings to a stylesheet.

        _root is an internal parameter as this function is recursive."""

        # build up root part of pathname to reference
        if _root is None:
            path = []
            obj = self
            while not obj.parent.iswidget:
                path.insert(0, obj.name)
                obj = obj.parent
            path = ['', 'StyleSheet', obj.parent.typename] + path + ['']
            _root = '/'.join(path)

        # iterate over subsettings
        for name, setn in self.setdict.items():
            thispath = _root + name
            if isinstance(setn, Settings):
                # call recursively if this is a Settings
                setn.linkToStylesheet(_root=thispath+'/')
            elif not setn.hidden and (setn.isReference() or setn.isDefault()):
                # link to stylesheet if the setting is a visible one

                # check that reference resolves
                ref = Reference(thispath)
                try:
                    ref.resolve(setn)
                except Reference.ResolveException:
                    # leave it as it was
                    pass
                else:
                    if setn.isReference() and setn.getReference().split[0] == '..':
                        # convert a relative path to multiple references
                        paths = [thispath, setn.getReference().value]
                        ref = ReferenceMultiple(paths)

                    setn.set(ref)
                    setn.default = ref
